# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ("Printer", "make_colorama_printer", "make_default_printer", "make_simple_printer")

import sys
from collections.abc import Callable, Sequence
from typing import Generic, TextIO

from ._layout import _K, Layout, LayoutRow

_CHAR_DECOMPOSITION = {
    # This mapping provides the "logic" for how to decompose the relevant
    # box-drawing symbols into symbols with just a single line segment,
    # allowing us to do set-operations on just the single-segment characters.
    " ": frozenset(),
    "╴": frozenset({"╴"}),
    "╵": frozenset({"╵"}),
    "╶": frozenset({"╶"}),
    "╷": frozenset({"╷"}),
    "╯": frozenset({"╴", "╵"}),
    "─": frozenset({"╴", "╶"}),
    "╮": frozenset({"╴", "╷"}),
    "╰": frozenset({"╵", "╶"}),
    "│": frozenset({"╵", "╷"}),
    "╭": frozenset({"╶", "╷"}),
    "┴": frozenset({"╴", "╵", "╶"}),
    "┤": frozenset({"╴", "╵", "╷"}),
    "┬": frozenset({"╴", "╶", "╷"}),
    "├": frozenset({"╵", "╶", "╷"}),
    "┼": frozenset({"╴", "╵", "╶", "╷"}),
}

_CHAR_COMPOSITION = {v: k for k, v in _CHAR_DECOMPOSITION.items()}


class PrintRow:
    """Base class and default implementation for drawing a single row of a text
    DAG visualization.

    Parameters
    ----------
    width : `int`
        Number of columns the graph will occupy (not including text
        descriptions on the right).
    pad : `str`
        Character to use for empty cells.
    """

    def __init__(self, width: int, pad: str):
        self._cells = [pad] * width

    def set(self, x: int, char: str, style: tuple[str, str] = ("", "")) -> None:
        """Set the character in a single cell, overriding what is there now.

        Parameters
        ----------
        x : `int`
            Column index.
        char : `str`
            Single character value to set.
        style : `tuple` [ `str`, `str` ], optional
            Styling prefix and suffix strings.  Ignored by the base class.
        """
        self._cells[x] = char

    def vert(self, x: int, style: tuple[str, str] = ("", "")) -> None:
        """Add a vertical line, taking into account the current contents of the
        cell.

        Parameters
        ----------
        x : `int`
            Column index.
        style : `tuple` [ `str`, `str` ], optional
            Styling prefix and suffix strings.  Ignored by the base class.

        Notes
        -----
        This merges the vertical line with any other character other than a
        complete complete horizontal segment, which it replaces to represent a
        non-intersection as a visual "hop", i.e. ``───`` becomes ``─│─``.
        """
        if self._cells[x] in (" ", "─"):
            self.set(x, "│", style)
        else:
            self.update(x, "│", style)

    def update(self, x: int, char: str, style: tuple[str, str] = ("", "")) -> None:
        """Combine a new line-drawing character with the line segments already
        present in a cell.

        Parameters
        ----------
        x : `int`
            Column index.
        char : `str`
            Single character value to add.
        style : `tuple` [ `str`, `str` ], optional
            Styling prefix and suffix strings.  Ignored by the base class.
        """
        self.set(x, _CHAR_COMPOSITION[_CHAR_DECOMPOSITION[char] | _CHAR_DECOMPOSITION[self._cells[x]]], style)

    def bend(
        self,
        start: int,
        stop: int,
        start_style: tuple[str, str] = ("", ""),
        stop_style: tuple[str, str] = ("", ""),
    ) -> None:
        """Draw a sideways-S bend representing a connection in one column from
        above to a different column below.

        Parameters
        ----------
        start : `int`
            Incoming column index (connecting from the previous line, above).
        stop : `int`
            Outgoing column index (connecting to the next line, below).
        start_style : `tuple` [ `str`, `str` ], optional
            Styling prefix and suffix strings.  Ignored by the base class.
        stop_style : `tuple` [ `str`, `str` ], optional
            Styling prefix and suffix strings.  Ignored by the base class.

        Notes
        -----
        If the incoming and outgoing columns are the same, this yields a
        vertical line.
        """
        if start < stop:
            self.update(start, "╰", start_style)
            self.update(stop, "╮", stop_style)
            for x in range(start + 1, stop):
                self.update(x, "─", stop_style)
        elif start > stop:
            self.update(start, "╯", start_style)
            self.update(stop, "╭", stop_style)
            for x in range(stop + 1, start):
                self.update(x, "─", stop_style)
        else:
            self.update(start, "│", start_style)

    def finish(self) -> str:
        """Return the line printer's internal state into a string."""
        return "".join(self._cells)


def _default_get_text(node: _K, x: int, style: tuple[str, str]) -> str:
    """Return the default text to associate with a node.

    This function is the default value for the ``get_text`` argument to
    `Printer`; see that for details.
    """
    return str(node)


def _default_get_symbol(node: _K, x: int) -> str:
    """Return the default symbol for a node.

    This function is the default value for the ``get_symbol`` argument to
    `Printer`; see that for details.
    """
    return "⬤"


def _default_get_style(node: _K, x: int) -> tuple[str, str]:
    """Get the default styling suffix/prefix strings.

    This function is the default value for the ``get_style`` argument to
    `Printer`; see that for details.
    """
    return "", ""


class Printer(Generic[_K]):
    """High-level tool for drawing a text-based DAG visualization.

    Parameters
    ----------
    layout_width : `int`
        Logical width of the layout.  Actual width of the graph is
        ``layout_width * 2 + 1`` to space out the nodes and make line
        intersections and non-intersections comprehensible.
    pad : `str`, optional
        Character to use for unpopulated cells.
    make_blank_row : `~collections.abc.Callable`
        Callback that returns a new `PrintRow` instance with no new cells
        populated.  Callback arguments are ``(self.width, self.pad)``.  This
        can return a specialization of `PrintRow` to add support for styling.
    get_text : `~collections.abc.Callable`
        Callback that returns the text description for a node.  Arguments
        are the node key, the column in which the node appears, and a
        style prefix/suffix 2-tuple.
    get_symbol : `~collections.abc.Callable`
        Callback that returns the symbol for a node.  Arguments are the node
        key, the column in which the node appears.
    get_style : `~collections.abc.Callable`
        Callback that returns a prefix/suffix style 2-tuple for a node.  Prefix
        and suffix values are strings, but are otherwise subclass-dependent;
        styles are ignored by the base class.
    """

    def __init__(
        self,
        layout_width: int,
        *,
        pad: str = " ",
        make_blank_row: Callable[[int, str], PrintRow] = PrintRow,
        get_text: Callable[[_K, int, tuple[str, str]], str] = _default_get_text,
        get_symbol: Callable[[_K, int], str] = _default_get_symbol,
        get_style: Callable[[_K, int], tuple[str, str]] = _default_get_style,
    ):
        self.width = layout_width * 2 + 1
        self.pad = pad
        self.make_blank_row = make_blank_row
        self.get_text = get_text
        self.get_symbol = get_symbol
        self.get_style = get_style

    def print_row(
        self,
        stream: TextIO,
        layout_row: LayoutRow[_K],
    ) -> None:
        """Print a single row of the DAG visualization to a file-like object.

        Parameters
        ----------
        stream : `TextIO`
            Output stream to write to.
        layout_row : `LayoutRow`
            Struct that indicates the columns in which nodes an edges should
            be drawn.
        """
        node_style = self.get_style(layout_row.node, layout_row.x)
        if layout_row.continuing or layout_row.connecting:
            print_row = self.make_blank_row(self.width, self.pad)
            for x, source in layout_row.connecting:
                print_row.bend(
                    2 * x,
                    2 * layout_row.x,
                    start_style=self.get_style(source, x),
                    stop_style=node_style,
                )
            for x, source, _ in layout_row.continuing:
                print_row.vert(2 * x, self.get_style(source, x))
            stream.write(print_row.finish())
        stream.write("\n")
        print_row = self.make_blank_row(self.width, self.pad)
        for x, source, _ in layout_row.continuing:
            print_row.vert(2 * x, self.get_style(source, x))
        print_row.set(2 * layout_row.x, self.get_symbol(layout_row.node, layout_row.x), node_style)
        stream.write(print_row.finish())
        stream.write(self.pad * 2)
        stream.write(self.get_text(layout_row.node, layout_row.x, node_style))
        stream.write("\n")

    def print(self, stream: TextIO, layout: Layout) -> None:
        """Print the DAG visualization to a file-like object.

        Parameters
        ----------
        stream : `TextIO`
            Output stream to write to.
        layout : `Layout`
            Struct that determines the rows and columns in which nodes and
            edges are drawn.
        """
        for layout_row in layout:
            self.print_row(stream, layout_row)


class TerminalPrintRow(PrintRow):
    """Specialization of `PrintRow` for interactive terminals.

    This adds support for styling (mostly colors) using terminal escape
    sequences.

    Parameters
    ----------
    width : `int`
        Number of columns the graph will occupy (not including text
        descriptions on the right).
    pad : `str`
        Character to use for empty cells.
    """

    def __init__(self, width: int, pad: str):
        super().__init__(width, pad)
        self._styles = [("", "")] * width

    def set(self, x: int, char: str, style: tuple[str, str] = ("", "")) -> None:
        super().set(x, char)
        self._styles[x] = style

    def finish(self) -> str:
        return "".join(f"{prefix}{char}{suffix}" for char, (prefix, suffix) in zip(self._cells, self._styles))


def make_colorama_printer(layout_width: int, palette: Sequence[str] = ()) -> Printer | None:
    """Return a `Printer` instance that uses terminal escape codes provided
    by the `colorama` package.

    Parameters
    ----------
    layout_width : `int`
        Logical width of the layout.  Actual width of the graph is
        ``layout_width * 2 + 1`` to space out the nodes and make line
        intersections and non-intersections comprehensible.
    palette : `~collections.abc.Sequence` [ `str` ], optional
        Sequence of colors, in which each is a single character (any of
        ``rgbcym``), a color name (any of ``red``, ``green``, ``blue``,
        ``cyan``, ``yellow``, or ``magenta``), or one of those color names
        preceded by ``light`` (with no space).  Case is ignored.  If empty,
        a predefined sequence using both light and dark colors is used.

    Returns
    -------
    printer : `Printer` or `None`
        A `Printer` instance, or `None` if the `colorama` package could not
        be imported.
    """
    try:
        import colorama
    except ImportError:
        return None
    if not palette:
        palette = [
            colorama.Fore.RED,
            colorama.Fore.LIGHTBLUE_EX,
            colorama.Fore.GREEN,
            colorama.Fore.LIGHTMAGENTA_EX,
            colorama.Fore.YELLOW,
            colorama.Fore.LIGHTCYAN_EX,
            colorama.Fore.LIGHTRED_EX,
            colorama.Fore.BLUE,
            colorama.Fore.LIGHTGREEN_EX,
            colorama.Fore.MAGENTA,
            colorama.Fore.LIGHTYELLOW_EX,
            colorama.Fore.CYAN,
        ]
    else:
        translate_color = {
            "R": colorama.Fore.RED,
            "RED": colorama.Fore.RED,
            "LIGHTRED": colorama.Fore.LIGHTRED_EX,
            "G": colorama.Fore.GREEN,
            "GREEN": colorama.Fore.GREEN,
            "LIGHTGREEN": colorama.Fore.LIGHTGREEN_EX,
            "B": colorama.Fore.BLUE,
            "BLUE": colorama.Fore.BLUE,
            "LIGHTBLUE": colorama.Fore.LIGHTBLUE_EX,
            "C": colorama.Fore.CYAN,
            "CYAN": colorama.Fore.CYAN,
            "LIGHTCYAN": colorama.Fore.LIGHTCYAN_EX,
            "Y": colorama.Fore.YELLOW,
            "YELLOW": colorama.Fore.YELLOW,
            "LIGHTYELLOW": colorama.Fore.LIGHTYELLOW_EX,
            "M": colorama.Fore.MAGENTA,
            "MAGENTA": colorama.Fore.MAGENTA,
            "LIGHTMAGENTA": colorama.Fore.LIGHTMAGENTA_EX,
        }
        palette = [translate_color.get(c.upper(), c) for c in palette]
    return Printer(
        layout_width,
        make_blank_row=lambda width, pad: TerminalPrintRow(width, pad),
        get_style=lambda node, x: (palette[x % len(palette)], colorama.Style.RESET_ALL),
    )


def make_simple_printer(layout_width: int) -> Printer:
    """Return a simple `Printer` instance with no styling.

    Parameters
    ----------
    layout_width : `int`
        Logical width of the layout.  Actual width of the graph is
        ``layout_width * 2 + 1`` to space out the nodes and make line
        intersections and non-intersections comprehensible.
    """
    return Printer(layout_width)


def make_default_printer(
    layout_width: int, color: bool | Sequence[str] | None = None, stream: TextIO = sys.stdout
) -> Printer:
    """Return a `Printer` instance with terminal escape-code coloring if
    possible (and requested), or a simple one otherwise.

    Parameters
    ----------
    layout_width : `int`
        Logical width of the layout.  Actual width of the graph is
        ``layout_width * 2 + 1`` to space out the nodes and make line
        intersections and non-intersections comprehensible.
    color : `bool` or `~collections.abc.Sequence` [ `str` ]
        Whether to use terminal escape codes to add color to the graph. Default
        is to add colors only if the `colorama` package can be imported.
        `False` disables colors unconditionally, while `True` or a sequence of
        colors (see `make_colorama_printer`) will result in `ImportError` being
        propagated up if `colorama` is unavailable.
    stream : `io.TextIO`, optional
        Output stream the printer will write to.
    """
    if color is None:
        if stream.isatty():
            if printer := make_colorama_printer(layout_width):
                return printer
    elif color:
        palette = color if color is not True else ()
        printer = make_colorama_printer(layout_width, palette)
        if printer is None:
            raise ImportError("Cannot use color unless the 'colorama' module is available.")
        return printer
    return make_simple_printer(layout_width)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ["filter_by_dataset_type_glob", "filter_by_existence"]

import re
from collections.abc import Collection

from lsst.daf.butler import Butler, DatasetRef
from lsst.daf.butler.utils import globToRegex
from lsst.utils.logging import getLogger
from lsst.utils.timer import time_this

_LOG = getLogger(__name__)


def _matches_dataset_type(dataset_type_name: str, regexes: list[str | re.Pattern]) -> bool:
    for regex in regexes:
        if isinstance(regex, str):
            if dataset_type_name == regex:
                return True
        elif regex.search(dataset_type_name):
            return True
    return False


def filter_by_dataset_type_glob(
    refs: Collection[DatasetRef], dataset_types: tuple[str, ...]
) -> Collection[DatasetRef]:
    """Filter the refs based on requested dataset types.

    Parameters
    ----------
    refs : `collections.abc.Collection` [ `lsst.daf.butler.DatasetRef` ]
        Datasets to be filtered.
    dataset_types : `tuple` [ `str`, ...]
        Dataset type names or globs to use for filtering. Empty tuple implies
        no filtering.

    Returns
    -------
    filtered : `collections.abc.Collection` [ `lsst.daf.butler.DatasetRef` ]
        Filter datasets.
    """
    regexes = globToRegex(dataset_types)
    if regexes is ...:
        # Nothing to do.
        return refs

    return {ref for ref in refs if _matches_dataset_type(ref.datasetType.name, regexes)}


def filter_by_existence(butler: Butler, refs: Collection[DatasetRef]) -> Collection[DatasetRef]:
    """Filter out the refs that the butler already knows exist.

    Parameters
    ----------
    butler : `lsst.daf.butler.Butler`
        Butler in which to check existence of given datarefs.
    refs : `collections.abc.Collection` [ `lsst.daf.butler.DatasetRef` ]
        Datasets to be filtered.

    Returns
    -------
    filtered : `collections.abc.Collection` [ `lsst.daf.butler.DatasetRef` ]
        Filter datasets.
    """
    _LOG.verbose("Filtering out datasets already known to the target butler...")
    with time_this(log=_LOG, msg="Completed checking existence"):
        existence = butler._datastore.knows_these(refs)
        filtered = [ref for ref in existence if not existence[ref]]
    _LOG.verbose(
        "After filtering out those already in the target butler, number of datasets to transfer: %d",
        len(filtered),
    )

    return filtered



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ()

import operator
import unittest
from collections import defaultdict

import lsst.pipe.base.connectionTypes as cT
from lsst.daf.butler import Butler, DataCoordinate
from lsst.daf.butler.tests.utils import create_populated_sqlite_registry
from lsst.pipe.base import (
    PipelineGraph,
    PipelineTask,
    PipelineTaskConfig,
    PipelineTaskConnections,
    QuantaAdjuster,
)
from lsst.pipe.base.all_dimensions_quantum_graph_builder import AllDimensionsQuantumGraphBuilder
from lsst.resources import ResourcePath


class GroupTestConnections(PipelineTaskConnections, dimensions=("detector",)):
    """Connections for a task whose quanta read in all of the biases for all
    detectors with the same purpose, use flat-field exposures as prerequisites,
    and (theoretically) writes both a single summary output for all inputs and
    another output for each input. The data IDs of the quanta and the summary
    outputs using the detector with the lowest ID in that group.
    """

    input_group = cT.Input(
        "bias",
        "Exposure",
        multiple=True,
        dimensions=("detector",),
        isCalibration=True,
    )
    prereq_input_group = cT.PrerequisiteInput(
        "flat",
        "Exposure",
        multiple=True,
        dimensions=("detector", "physical_filter", "band"),
        isCalibration=True,
    )
    output_group = cT.Output(
        "bias_stuff",
        "StructuredDataDict",
        multiple=True,
        dimensions=("detector",),
    )
    single_output = cT.Output(
        "bias_summary",
        "StructuredDataDict",
        multiple=False,
        dimensions=("detector",),
    )

    def adjust_all_quanta(self, adjuster: QuantaAdjuster) -> None:
        # Group the quanta by their detector's purpose.
        quanta_by_detector_purpose: defaultdict[str, list[DataCoordinate]] = defaultdict(list)
        for quantum_data_id in adjuster.iter_data_ids():
            quantum_data_id = adjuster.expand_quantum_data_id(quantum_data_id)
            quanta_by_detector_purpose[quantum_data_id.records["detector"].purpose].append(quantum_data_id)
        # Within each group, keep only the one with the lowest detector ID,
        # while transferring the inputs and outputs of the others to that
        # quantum.
        for data_id_group in quanta_by_detector_purpose.values():
            data_id_group.sort(key=operator.itemgetter("detector"))
            keep, *drop = data_id_group
            for drop_data_id in drop:
                for input_data_id in adjuster.get_inputs(drop_data_id)["input_group"]:
                    adjuster.add_input(keep, "input_group", input_data_id)
                for input_uuid in adjuster.get_prerequisite_inputs(drop_data_id)["prereq_input_group"]:
                    adjuster.add_prerequisite_input(keep, "prereq_input_group", input_uuid)
                for output_data_id in adjuster.get_outputs(drop_data_id)["output_group"]:
                    adjuster.move_output(keep, "output_group", output_data_id)
                adjuster.remove_quantum(drop_data_id)


class GroupTestConfig(PipelineTaskConfig, pipelineConnections=GroupTestConnections):
    pass


class GroupTestTask(PipelineTask):
    ConfigClass = GroupTestConfig


class AdjustAllQuantaTestCase(unittest.TestCase):
    """Tests for the `PipelineTaskConnections.adjust_all_quanta` hook in
    quantum-graph generation.
    """

    @staticmethod
    def make_butler() -> Butler:
        DATA_ROOT = ResourcePath("resource://lsst.daf.butler/tests/registry_data", forceDirectory=True)
        return create_populated_sqlite_registry(
            *[DATA_ROOT.join(filename) for filename in ("base.yaml", "datasets.yaml")]
        )

    def test_adjust_all_quanta(self) -> None:
        """Build a quantum graph for a task that implements the
        adjust_all_quanta hook, and check that it works as expected.
        """
        butler = self.make_butler()
        pipeline_graph = PipelineGraph(universe=butler.dimensions)
        pipeline_graph.add_task("grouper", GroupTestTask)
        collections = ["imported_g", "imported_r"]
        qgb = AllDimensionsQuantumGraphBuilder(
            pipeline_graph,
            butler,
            input_collections=collections,
            output_run="irrelevant",
        )
        qg = qgb.build(attach_datastore_records=False)
        quanta = {quantum.dataId["detector"]: quantum for quantum in qg.get_task_quanta("grouper").values()}
        # This test camera (defined in daf_butler test data) has 4 detectors;
        # 1-3 have purpose=SCIENCE, and 4 has purpose=WAVEFRONT.
        self.assertEqual(quanta.keys(), {1, 4})
        self.assertEqual(len(quanta[1].inputs["bias"]), 3)
        self.assertEqual(len(quanta[1].inputs["flat"]), 5)
        self.assertEqual(len(quanta[1].outputs["bias_stuff"]), 3)
        self.assertCountEqual(
            quanta[1].inputs["bias"],
            butler.query_datasets("bias", collections=collections, where="detector.purpose = 'SCIENCE'"),
        )
        self.assertCountEqual(
            quanta[1].inputs["flat"],
            butler.query_datasets("flat", collections=collections, where="detector.purpose = 'SCIENCE'"),
        )
        self.assertEqual(len(quanta[1].outputs["bias_summary"]), 1)
        self.assertEqual(quanta[1].outputs["bias_summary"][0].dataId["detector"], 1)
        self.assertEqual(len(quanta[4].inputs["bias"]), 1)
        self.assertEqual(len(quanta[4].inputs["flat"]), 2)
        self.assertEqual(len(quanta[4].outputs["bias_stuff"]), 1)
        self.assertCountEqual(
            quanta[4].inputs["bias"],
            butler.query_datasets("bias", collections=collections, where="detector.purpose = 'WAVEFRONT'"),
        )
        self.assertCountEqual(
            quanta[4].inputs["flat"],
            butler.query_datasets("flat", collections=collections, where="detector.purpose = 'WAVEFRONT'"),
        )
        self.assertEqual(len(quanta[4].outputs["bias_summary"]), 1)
        self.assertEqual(quanta[4].outputs["bias_summary"][0].dataId["detector"], 4)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ["InMemoryLimitedButler"]

import logging
import uuid
from collections.abc import Iterable
from typing import Any

from lsst.daf.butler import (
    ButlerMetrics,
    DatasetProvenance,
    DatasetRef,
    DatasetType,
    DimensionUniverse,
    LimitedButler,
    MissingDatasetTypeError,
    Quantum,
    StorageClass,
    StorageClassFactory,
)
from lsst.daf.butler.registry import ConflictingDefinitionError

from .._dataset_handle import InMemoryDatasetHandle

_LOG = logging.getLogger(__name__)


class InMemoryLimitedButler(LimitedButler):
    """A `LimitedButler` that just stores datasets in an in-memory mapping.

    Parameters
    ----------
    universe : `lsst.daf.butler.DimensionUniverse`
        Definitions for all dimensions.
    dataset_types : `~collections.abc.Iterable` [ \
            `lsst.daf.butler.DatasetType` ]
        Definitions of all dataset types.

    Notes
    -----
    This is an incomplete implementation of the `LimitedButler` interface
    intended only for tests.  It supports all methods required by
    `SingleQuantumExecutor`, but not transfers or URI retrieval.

    While this class supports storage class conversions in `get` and `put`, it
    uses different code paths from real butlers, and should not be used in
    tests in which storage class correctness is part of what is being tested.

    Objects are always copied (via storage class machinery) by `get`.

    Pickling this class will pickle all datasets already `put` (which must be
    pickleable).  This generally allows a central butler to be initialized with
    input datasets in one process and distributed to worker processes that run
    quanta *once*, but it does not allow outputs from a worker process to be
    distributed to others or the originating process.  This can be hard to
    notice because quanta will usually be skipped with
    `lsst.pipe.base.NoWorkFound` (a success!) when all of their inputs are
    missing.
    """

    def __init__(self, universe: DimensionUniverse, dataset_types: Iterable[DatasetType] = ()):
        self.storageClasses = StorageClassFactory()
        self._universe = universe
        self._datasets: dict[uuid.UUID, tuple[DatasetRef, InMemoryDatasetHandle]] = {}
        self._metrics = ButlerMetrics()
        self._dataset_types = {dt.name: dt for dt in dataset_types}
        assert not any(dt.component() for dt in self._dataset_types.values()), (
            "Dataset type definitions must not be components."
        )

    def __getstate__(self) -> dict[str, Any]:
        # Pickle customization is needed because StorageClassFactory is not
        # pickleable.
        return {
            "universe": self._universe,
            "datasets": self._datasets,
            "dataset_types": self._dataset_types,
        }

    def __setstate__(self, state: dict[str, Any]) -> None:
        self.storageClasses = StorageClassFactory()
        self._universe = state["universe"]
        self._datasets = state["datasets"]
        self._metrics = ButlerMetrics()
        self._dataset_types = state["dataset_types"]

    def get_datasets(self, dataset_type: str | None = None) -> dict[DatasetRef, object]:
        """Return datasets that have been `put` to this butler.

        Storage classes and corresponding Python types will match the dataset
        type definitions provided at butler construction, which may not be the
        same as what was `put`.

        Parameters
        ----------
        dataset_type : `str`, optional
            Dataset type name used to filter results.

        Returns
        -------
        refs : `dict` [ `lsst.daf.butler.DatasetRef`, `object` ]
            Datasets held by this butler.
        """
        return {
            ref: handle.get()
            for ref, handle in self._datasets.values()
            if dataset_type is None or dataset_type == ref.datasetType.name
        }

    def isWriteable(self) -> bool:
        return True

    def put(self, obj: Any, ref: DatasetRef, /, *, provenance: DatasetProvenance | None = None) -> DatasetRef:
        with self._metrics.instrument_put():
            assert not ref.isComponent(), "Component dataset types cannot be put."
            if ref.id in self._datasets:
                # Some butlers may not raise reliably when a dataset already
                # exists (it's hard to be rigorous in parallel given different
                # guarantees provided by storage), but we don't want code to
                # rely on it not being an error, so we want a test butler to
                # always complain.
                raise ConflictingDefinitionError(f"Dataset {ref} already exists.")
            if (repo_dataset_type := self._dataset_types.get(ref.datasetType.name)) is None:
                raise MissingDatasetTypeError(f"Dataset type {ref.datasetType.name!r} not recognized.")
            repo_dataset_type.storageClass.coerce_type(obj)
            self._datasets[ref.id] = (
                ref.overrideStorageClass(repo_dataset_type.storageClass),
                InMemoryDatasetHandle(
                    obj,
                    storageClass=repo_dataset_type.storageClass,
                    dataId=ref.dataId,
                    copy=True,
                ),
            )
        return ref

    def get(
        self,
        ref: DatasetRef,
        /,
        *,
        parameters: dict[str, Any] | None = None,
        storageClass: StorageClass | str | None = None,
    ) -> Any:
        with self._metrics.instrument_get():
            if storageClass is None:
                storageClass = ref.datasetType.storageClass
            elif isinstance(storageClass, str):
                storageClass = self.storageClasses.getStorageClass(storageClass)
            if entry := self._datasets.get(ref.id):
                (ref, handle) = entry
                return handle.get(
                    component=ref.datasetType.component(), parameters=parameters, storageClass=storageClass
                )
        raise FileNotFoundError(f"Dataset {ref} does not exist.")

    def stored_many(self, refs: Iterable[DatasetRef]) -> dict[DatasetRef, bool]:
        return {ref: ref.id in self._datasets for ref in refs}

    def pruneDatasets(
        self,
        refs: Iterable[DatasetRef],
        *,
        disassociate: bool = True,
        unstore: bool = False,
        tags: Iterable[str] = (),
        purge: bool = False,
    ) -> None:
        for ref in refs:
            self._datasets.pop(ref.id, None)

    @property
    def _datastore(self) -> Any:
        raise NotImplementedError("This test butler does not have a datastore.")

    @_datastore.setter  # demanded by MyPy since we declare it to be an instance attribute in LimitedButler.
    def _datastore(self, value: Any) -> None:
        raise NotImplementedError("This test butler does not have a datastore.")

    @property
    def dimensions(self) -> DimensionUniverse:
        return self._universe

    def factory(self, quantum: Quantum) -> InMemoryLimitedButler:
        """Return ``self``.

        This method can be used as the ``limited_butler_factory`` argument to
        `.single_quantum_executor.SingleQuantumExecutor`.

        Parameters
        ----------
        quantum : `lsst.daf.butler.Quantum`
            Ignored.
        """
        return self



# # This file is part of pipe_base.
# #
# # Developed for the LSST Data Management System.
# # This product includes software developed by the LSST Project
# # (http://www.lsst.org).
# # See the COPYRIGHT file at the top-level directory of this distribution
# # for details of code ownership.
# #
# # This software is dual licensed under the GNU General Public License and
# also
# # under a 3-clause BSD license. Recipients may choose which of these licenses
# # to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# # respectively.  If you choose the GPL option then the following text applies
# # (but note that there is still no warranty even if you opt for BSD instead):
# #
# # This program is free software: you can redistribute it and/or modify
# # it under the terms of the GNU General Public License as published by
# # the Free Software Foundation, either version 3 of the License, or
# # (at your option) any later version.
# #
# # This program is distributed in the hope that it will be useful,
# # but WITHOUT ANY WARRANTY; without even the implied warranty of
# # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# # GNU General Public License for more details.
# #
# # You should have received a copy of the GNU General Public License
# # along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Simple unit test for quantum_provenance_graph."""

import unittest
import uuid

import lsst.utils.logging
from lsst.pipe.base.quantum_provenance_graph import (
    CursedDatasetSummary,
    DatasetTypeSummary,
    ExceptionInfo,
    ExceptionInfoSummary,
    QuantumProvenanceGraph,
    Summary,
    TaskSummary,
    UnsuccessfulQuantumSummary,
)
from lsst.pipe.base.tests import simpleQGraph
from lsst.utils.tests import temporaryDirectory

expected_mock_datasets = [
    "add_dataset1",
    "add2_dataset1",
    "task0_metadata",
    "task0_log",
    "add_dataset2",
    "add2_dataset2",
    "task1_metadata",
    "task1_log",
    "add_dataset3",
    "add2_dataset3",
    "task2_metadata",
    "task2_log",
    "add_dataset4",
    "add2_dataset4",
    "task3_metadata",
    "task3_log",
    "add_dataset5",
    "add2_dataset5",
    "task4_metadata",
    "task4_log",
]


class QuantumProvenanceGraphTestCase(unittest.TestCase):
    """Test reports from the QuantumProvenanceGraph.

    Verify that the `QuantumProvenanceGraph` is able to extract correct
    information from `simpleQgraph`.

    More tests are in lsst/ci_middleware/tests/test_prod_outputs.py and
    lsst/ci_middleware/tests/test_rc2_outputs.py
    """

    def test_qpg_reports(self) -> None:
        """Test that we can add a new graph to the
        `QuantumProvenanceGraph`.
        """
        with temporaryDirectory() as root:
            # make a simple qgraph to make an execution report on
            butler, qgraph = simpleQGraph.makeSimpleQGraph(root=root)
            qpg = QuantumProvenanceGraph(butler, [qgraph])
            summary = qpg.to_summary(butler)

            for task_summary in summary.tasks.values():
                # We know that we have one expected task that was not run.
                # As such, the following dictionary should describe all of
                # the mock tasks.
                self.assertEqual(task_summary.n_successful, 0)
                self.assertEqual(task_summary.n_blocked, 0)
                self.assertEqual(task_summary.n_unknown, 1)
                self.assertEqual(task_summary.n_expected, 1)
                self.assertListEqual(task_summary.failed_quanta, [])
                self.assertListEqual(task_summary.recovered_quanta, [])
                self.assertListEqual(task_summary.wonky_quanta, [])
                self.assertDictEqual(task_summary.exceptions, {})
                self.assertEqual(task_summary.n_wonky, 0)
                self.assertEqual(task_summary.n_failed, 0)

            for dataset_type_name, dataset_type_summary in summary.datasets.items():
                self.assertListEqual(
                    dataset_type_summary.unsuccessful_datasets,
                    [{"instrument": "INSTR", "detector": 0}],
                )
                # Check dataset counts (can't be done all in one because
                # datasets have different producers), but all the counts for
                # each task should be the same.
                self.assertEqual(dataset_type_summary.n_visible, 0)
                self.assertEqual(dataset_type_summary.n_shadowed, 0)
                self.assertEqual(dataset_type_summary.n_predicted_only, 0)
                self.assertEqual(dataset_type_summary.n_expected, 1)
                self.assertEqual(dataset_type_summary.n_cursed, 0)
                self.assertEqual(dataset_type_summary.n_unsuccessful, 1)
                # Make sure the cursed dataset is an empty list
                self.assertListEqual(dataset_type_summary.cursed_datasets, [])
                # Make sure we have the right datasets based on our mock
                self.assertIn(dataset_type_name, expected_mock_datasets)
                # Make sure the expected datasets were produced by the expected
                # tasks
                match dataset_type_name:
                    case name if name in ["add_dataset1", "add2_dataset1", "task0_metadata", "task0_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task0")
                    case name if name in ["add_dataset2", "add2_dataset2", "task1_metadata", "task1_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task1")
                    case name if name in ["add_dataset3", "add2_dataset3", "task2_metadata", "task2_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task2")
                    case name if name in ["add_dataset4", "add2_dataset4", "task3_metadata", "task3_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task3")
                    case name if name in ["add_dataset5", "add2_dataset5", "task4_metadata", "task4_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task4")

    def test_aggregate_reports(self) -> None:
        """Test aggregating reports from the `QuantumProvenanceGraph.`"""
        with temporaryDirectory() as root:
            # make a simple qgraph to make an execution report on
            butler, qgraph = simpleQGraph.makeSimpleQGraph(root=root)
            qpg = QuantumProvenanceGraph(butler, [qgraph])
            summary = qpg.to_summary(butler)
            # Check that aggregating one summary only does not cause an error
            one_graph_only_sum = Summary.aggregate([summary])

            # Do the same tests as in `test_qpg_reports`, but on the
            # 'aggregate' summary. Essentially, verify that the information in
            # the report is preserved during the aggregation step.
            for task_summary in one_graph_only_sum.tasks.values():
                self.assertEqual(task_summary.n_successful, 0)
                self.assertEqual(task_summary.n_blocked, 0)
                self.assertEqual(task_summary.n_unknown, 1)
                self.assertEqual(task_summary.n_expected, 1)
                self.assertListEqual(task_summary.failed_quanta, [])
                self.assertListEqual(task_summary.recovered_quanta, [])
                self.assertListEqual(task_summary.wonky_quanta, [])
                self.assertDictEqual(task_summary.exceptions, {})
                self.assertEqual(task_summary.n_wonky, 0)
                self.assertEqual(task_summary.n_failed, 0)
            for dataset_type_name, dataset_type_summary in one_graph_only_sum.datasets.items():
                self.assertListEqual(
                    dataset_type_summary.unsuccessful_datasets, [{"instrument": "INSTR", "detector": 0}]
                )
                self.assertEqual(dataset_type_summary.n_visible, 0)
                self.assertEqual(dataset_type_summary.n_shadowed, 0)
                self.assertEqual(dataset_type_summary.n_predicted_only, 0)
                self.assertEqual(dataset_type_summary.n_expected, 1)
                self.assertEqual(dataset_type_summary.n_cursed, 0)
                self.assertEqual(dataset_type_summary.n_unsuccessful, 1)
                self.assertListEqual(dataset_type_summary.cursed_datasets, [])
                self.assertIn(dataset_type_name, expected_mock_datasets)
                match dataset_type_name:
                    case name if name in ["add_dataset1", "add2_dataset1", "task0_metadata", "task0_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task0")
                    case name if name in ["add_dataset2", "add2_dataset2", "task1_metadata", "task1_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task1")
                    case name if name in ["add_dataset3", "add2_dataset3", "task2_metadata", "task2_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task2")
                    case name if name in ["add_dataset4", "add2_dataset4", "task3_metadata", "task3_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task3")
                    case name if name in ["add_dataset5", "add2_dataset5", "task4_metadata", "task4_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task4")

            # Now we test aggregating multiple summaries. First, we try
            # aggregating with an exact copy and make sure we just have double
            # the numbers.
            summary2 = summary.model_copy(deep=True)
            two_identical_graph_sum = Summary.aggregate([summary, summary2])
            for task_summary in two_identical_graph_sum.tasks.values():
                self.assertEqual(task_summary.n_successful, 0)
                self.assertEqual(task_summary.n_blocked, 0)
                self.assertEqual(task_summary.n_unknown, 2)
                self.assertEqual(task_summary.n_expected, 2)
                self.assertListEqual(task_summary.failed_quanta, [])
                self.assertListEqual(task_summary.recovered_quanta, [])
                self.assertListEqual(task_summary.wonky_quanta, [])
                self.assertDictEqual(task_summary.exceptions, {})
                self.assertEqual(task_summary.n_wonky, 0)
                self.assertEqual(task_summary.n_failed, 0)
            for dataset_type_name, dataset_type_summary in two_identical_graph_sum.datasets.items():
                self.assertListEqual(
                    dataset_type_summary.unsuccessful_datasets,
                    [{"instrument": "INSTR", "detector": 0}, {"instrument": "INSTR", "detector": 0}],
                )
                self.assertEqual(dataset_type_summary.n_visible, 0)
                self.assertEqual(dataset_type_summary.n_shadowed, 0)
                self.assertEqual(dataset_type_summary.n_predicted_only, 0)
                self.assertEqual(dataset_type_summary.n_expected, 2)
                self.assertEqual(dataset_type_summary.n_cursed, 0)
                self.assertEqual(dataset_type_summary.n_unsuccessful, 2)
                self.assertListEqual(dataset_type_summary.cursed_datasets, [])
                self.assertIn(dataset_type_name, expected_mock_datasets)
                match dataset_type_name:
                    case name if name in ["add_dataset1", "add2_dataset1", "task0_metadata", "task0_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task0")
                    case name if name in ["add_dataset2", "add2_dataset2", "task1_metadata", "task1_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task1")
                    case name if name in ["add_dataset3", "add2_dataset3", "task2_metadata", "task2_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task2")
                    case name if name in ["add_dataset4", "add2_dataset4", "task3_metadata", "task3_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task3")
                    case name if name in ["add_dataset5", "add2_dataset5", "task4_metadata", "task4_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task4")

            # Let's see if we can change lots of counts and info for a task
            # which exists in summary 1 and append to the overall summary
            # effectively. This summary has a lot of valid variations.
            uuid_a = uuid.uuid4()
            summary3 = summary.model_copy(deep=True)
            summary3.tasks["task4"] = TaskSummary.model_validate(
                {
                    "n_successful": 10,
                    "n_blocked": 20,
                    "n_unknown": 4,
                    "n_expected": 47,
                    "failed_quanta": [
                        {
                            "data_id": {"instrument": "INSTR", "detector": 1},
                            "runs": {"run1": "failed"},
                            "messages": ["Error on detector 1", "Second error on detector 1"],
                        },
                        {
                            "data_id": {"instrument": "INSTR", "detector": 2},
                            "runs": {"run1": "failed", "run2": "failed"},
                            "messages": [],
                        },
                        {
                            "data_id": {"instrument": "INSTR", "detector": 3},
                            "runs": {"run1": "failed"},
                            "messages": ["Error on detector 3"],
                        },
                    ],
                    "recovered_quanta": [
                        {"instrument": "INSTR", "detector": 4},
                        {"instrument": "INSTR", "detector": 5},
                        {"instrument": "INSTR", "detector": 6},
                    ],
                    "wonky_quanta": [
                        {
                            "data_id": {"instrument": "INSTR", "detector": 7},
                            "runs": {"run1": "successful", "run2": "failed"},
                            "messages": ["This one is wonky because it moved from successful to failed."],
                        }
                    ],
                    "caveats": {
                        "+A": [{"instrument": "INSTR", "detector": 10}],
                    },
                    "exceptions": {
                        "lsst.pipe.base.tests.mocks.MockAlgorithmError": [
                            {
                                "quantum_id": str(uuid_a),
                                "data_id": {"instrument": "INSTR", "detector": 10},
                                "run": "run2",
                                "exception": {
                                    "type_name": "lsst.pipe.base.tests.mocks.MockAlgorithmError",
                                    "message": "message A",
                                    "metadata": {"badness": 11},
                                },
                            }
                        ],
                    },
                    "n_wonky": 1,
                    "n_failed": 3,
                }
            )
            summary3.datasets["add_dataset5"] = DatasetTypeSummary.model_validate(
                {
                    "producer": "task4",
                    "n_visible": 0,
                    "n_shadowed": 0,
                    "n_predicted_only": 0,
                    "n_expected": 47,
                    "cursed_datasets": [
                        {
                            "producer_data_id": {"instrument": "INSTR", "detector": 7},
                            "data_id": {"instrument": "INSTR", "detector": 7},
                            "runs_produced": {"run1": True, "run2": False},
                            "run_visible": None,
                            "messages": ["Some kind of cursed dataset."],
                        }
                    ],
                    "unsuccessful_datasets": [
                        {"instrument": "INSTR", "detector": 0},
                        {"instrument": "INSTR", "detector": 1},
                        {"instrument": "INSTR", "detector": 2},
                        {"instrument": "INSTR", "detector": 3},
                    ],
                }
            )
            summary3.datasets["add2_dataset5"] = DatasetTypeSummary.model_validate(
                {
                    "producer": "task4",
                    "n_visible": 0,
                    "n_shadowed": 0,
                    "n_predicted_only": 0,
                    "n_expected": 47,
                    "cursed_datasets": [
                        {
                            "producer_data_id": {"instrument": "INSTR", "detector": 7},
                            "data_id": {"instrument": "INSTR", "detector": 7},
                            "runs_produced": {"run1": True, "run2": False},
                            "run_visible": None,
                            "messages": ["Some kind of cursed dataset."],
                        }
                    ],
                    "unsuccessful_datasets": [
                        {"instrument": "INSTR", "detector": 0},
                        {"instrument": "INSTR", "detector": 1},
                        {"instrument": "INSTR", "detector": 2},
                        {"instrument": "INSTR", "detector": 3},
                    ],
                }
            )
            # Test that aggregate with this file works
            two_graphs_different_numbers = Summary.aggregate([summary, summary3])
            for task_label, task_summary in two_graphs_different_numbers.tasks.items():
                if task_label == "task4":
                    self.assertEqual(task_summary.n_successful, 10)
                    self.assertEqual(task_summary.n_blocked, 20)
                    self.assertEqual(task_summary.n_unknown, 5)
                    self.assertEqual(task_summary.n_expected, 48)
                    self.assertListEqual(
                        task_summary.failed_quanta,
                        [
                            UnsuccessfulQuantumSummary(
                                data_id={"instrument": "INSTR", "detector": 1},
                                runs={"run1": "failed"},
                                messages=["Error on detector 1", "Second error on detector 1"],
                            ),
                            UnsuccessfulQuantumSummary(
                                data_id={"instrument": "INSTR", "detector": 2},
                                runs={"run1": "failed", "run2": "failed"},
                                messages=[],
                            ),
                            UnsuccessfulQuantumSummary(
                                data_id={"instrument": "INSTR", "detector": 3},
                                runs={"run1": "failed"},
                                messages=["Error on detector 3"],
                            ),
                        ],
                    )
                    self.assertListEqual(
                        task_summary.recovered_quanta,
                        [
                            {"instrument": "INSTR", "detector": 4},
                            {"instrument": "INSTR", "detector": 5},
                            {"instrument": "INSTR", "detector": 6},
                        ],
                    )
                    self.assertListEqual(
                        task_summary.wonky_quanta,
                        [
                            UnsuccessfulQuantumSummary(
                                data_id={"instrument": "INSTR", "detector": 7},
                                runs={"run1": "successful", "run2": "failed"},
                                messages=["This one is wonky because it moved from successful to failed."],
                            )
                        ],
                    )
                    self.assertDictEqual(
                        task_summary.caveats,
                        {"+A": [{"instrument": "INSTR", "detector": 10}]},
                    )
                    self.assertDictEqual(
                        task_summary.exceptions,
                        {
                            "lsst.pipe.base.tests.mocks.MockAlgorithmError": [
                                ExceptionInfoSummary(
                                    quantum_id=uuid_a,
                                    data_id={"instrument": "INSTR", "detector": 10},
                                    run="run2",
                                    exception=ExceptionInfo(
                                        type_name="lsst.pipe.base.tests.mocks.MockAlgorithmError",
                                        message="message A",
                                        metadata={"badness": 11},
                                    ),
                                )
                            ],
                        },
                    )
                    self.assertEqual(task_summary.n_wonky, 1)
                    self.assertEqual(task_summary.n_failed, 3)
                else:
                    self.assertEqual(task_summary.n_successful, 0)
                    self.assertEqual(task_summary.n_blocked, 0)
                    self.assertEqual(task_summary.n_unknown, 2)
                    self.assertEqual(task_summary.n_expected, 2)
                    self.assertListEqual(task_summary.failed_quanta, [])
                    self.assertListEqual(task_summary.recovered_quanta, [])
                    self.assertListEqual(task_summary.wonky_quanta, [])
                    self.assertEqual(task_summary.n_wonky, 0)
                    self.assertEqual(task_summary.n_failed, 0)
            for dataset_type_name, dataset_type_summary in two_graphs_different_numbers.datasets.items():
                if dataset_type_name in ["add_dataset5", "add2_dataset5"]:
                    self.assertListEqual(
                        dataset_type_summary.unsuccessful_datasets,
                        [
                            {"instrument": "INSTR", "detector": 0},
                            {"instrument": "INSTR", "detector": 0},
                            {"instrument": "INSTR", "detector": 1},
                            {"instrument": "INSTR", "detector": 2},
                            {"instrument": "INSTR", "detector": 3},
                        ],
                    )
                    self.assertEqual(dataset_type_summary.n_visible, 0)
                    self.assertEqual(dataset_type_summary.n_shadowed, 0)
                    self.assertEqual(dataset_type_summary.n_predicted_only, 0)
                    self.assertEqual(dataset_type_summary.n_expected, 48)
                    self.assertEqual(dataset_type_summary.n_cursed, 1)
                    self.assertEqual(dataset_type_summary.n_unsuccessful, 5)
                    self.assertListEqual(
                        dataset_type_summary.cursed_datasets,
                        [
                            CursedDatasetSummary(
                                producer_data_id={"instrument": "INSTR", "detector": 7},
                                data_id={"instrument": "INSTR", "detector": 7},
                                runs_produced={"run1": True, "run2": False},
                                run_visible=None,
                                messages=["Some kind of cursed dataset."],
                            )
                        ],
                    )
                else:
                    self.assertListEqual(
                        dataset_type_summary.unsuccessful_datasets,
                        [{"instrument": "INSTR", "detector": 0}, {"instrument": "INSTR", "detector": 0}],
                    )
                    self.assertEqual(dataset_type_summary.n_visible, 0)
                    self.assertEqual(dataset_type_summary.n_shadowed, 0)
                    self.assertEqual(dataset_type_summary.n_predicted_only, 0)
                    self.assertEqual(dataset_type_summary.n_expected, 2)
                    self.assertEqual(dataset_type_summary.n_cursed, 0)
                    self.assertEqual(dataset_type_summary.n_unsuccessful, 2)
                    self.assertListEqual(dataset_type_summary.cursed_datasets, [])
                    self.assertIn(dataset_type_name, expected_mock_datasets)
                match dataset_type_name:
                    case name if name in ["add_dataset1", "add2_dataset1", "task0_metadata", "task0_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task0")
                    case name if name in ["add_dataset2", "add2_dataset2", "task1_metadata", "task1_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task1")
                    case name if name in ["add_dataset3", "add2_dataset3", "task2_metadata", "task2_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task2")
                    case name if name in ["add_dataset4", "add2_dataset4", "task3_metadata", "task3_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task3")
                    case name if name in ["add_dataset5", "add2_dataset5", "task4_metadata", "task4_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task4")

            # Now, let's add a task to one model and see if aggregate still
            # works
            summary4 = summary.model_copy(deep=True)
            summary4.tasks["task5"] = TaskSummary(
                n_successful=0,
                n_blocked=0,
                n_unknown=1,
                n_expected=1,
                failed_quanta=[],
                recovered_quanta=[],
                wonky_quanta=[],
            )
            summary4.datasets["add_dataset6"] = DatasetTypeSummary(
                producer="task5",
                n_visible=0,
                n_shadowed=0,
                n_predicted_only=0,
                n_expected=1,
                cursed_datasets=[],
                unsuccessful_datasets=[{"instrument": "INSTR", "detector": 0}],
            )
            summary4.datasets["task5_log"] = DatasetTypeSummary(
                producer="task5",
                n_visible=0,
                n_shadowed=0,
                n_predicted_only=0,
                n_expected=1,
                cursed_datasets=[],
                unsuccessful_datasets=[{"instrument": "INSTR", "detector": 0}],
            )
            summary4.datasets["task5_metadata"] = DatasetTypeSummary(
                producer="task5",
                n_visible=0,
                n_shadowed=0,
                n_predicted_only=0,
                n_expected=1,
                cursed_datasets=[],
                unsuccessful_datasets=[{"instrument": "INSTR", "detector": 0}],
            )
            two_graphs_extra_task = Summary.aggregate([summary4, summary])
            # Make sure the extra task is in there
            self.assertIn("task5", two_graphs_extra_task.tasks)
            for task_label, task_summary in two_graphs_extra_task.tasks.items():
                self.assertEqual(task_summary.n_successful, 0)
                self.assertEqual(task_summary.n_blocked, 0)
                self.assertListEqual(task_summary.failed_quanta, [])
                self.assertListEqual(task_summary.recovered_quanta, [])
                self.assertListEqual(task_summary.wonky_quanta, [])
                self.assertEqual(task_summary.n_wonky, 0)
                self.assertEqual(task_summary.n_failed, 0)
                self.assertIn(task_label, ["task0", "task1", "task2", "task3", "task4", "task5"])
                if task_label == "task5":
                    self.assertEqual(task_summary.n_unknown, 1)
                    self.assertEqual(task_summary.n_expected, 1)
                else:
                    self.assertEqual(task_summary.n_unknown, 2)
                    self.assertEqual(task_summary.n_expected, 2)
            for dataset_type_name, dataset_type_summary in two_graphs_extra_task.datasets.items():
                self.assertEqual(dataset_type_summary.n_visible, 0)
                self.assertEqual(dataset_type_summary.n_shadowed, 0)
                self.assertEqual(dataset_type_summary.n_predicted_only, 0)
                self.assertEqual(dataset_type_summary.n_cursed, 0)
                self.assertListEqual(dataset_type_summary.cursed_datasets, [])
                if dataset_type_summary.producer == "task5":
                    self.assertEqual(dataset_type_summary.n_expected, 1)
                    self.assertEqual(dataset_type_summary.n_unsuccessful, 1)
                    self.assertListEqual(
                        dataset_type_summary.unsuccessful_datasets,
                        [{"instrument": "INSTR", "detector": 0}],
                    )
                else:
                    self.assertEqual(dataset_type_summary.n_expected, 2)
                    self.assertEqual(dataset_type_summary.n_unsuccessful, 2)
                    self.assertListEqual(
                        dataset_type_summary.unsuccessful_datasets,
                        [{"instrument": "INSTR", "detector": 0}, {"instrument": "INSTR", "detector": 0}],
                    )
                self.assertIn(
                    dataset_type_name,
                    expected_mock_datasets + ["add_dataset6", "task5_metadata", "task5_log"],
                )
                match dataset_type_name:
                    case name if name in ["add_dataset1", "add2_dataset1", "task0_metadata", "task0_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task0")
                    case name if name in ["add_dataset2", "add2_dataset2", "task1_metadata", "task1_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task1")
                    case name if name in ["add_dataset3", "add2_dataset3", "task2_metadata", "task2_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task2")
                    case name if name in ["add_dataset4", "add2_dataset4", "task3_metadata", "task3_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task3")
                    case name if name in ["add_dataset5", "add2_dataset5", "task4_metadata", "task4_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task4")
                    case name if name in ["add_dataset6", "task5_metadata", "task5_log"]:
                        self.assertEqual(dataset_type_summary.producer, "task5")

        # Now we test that we properly catch task-dataset mismatches in
        # aggregated graphs. This is a problem because if task 1 produced
        # a certain dataset in graph 1, but task 2 produced the same dataset
        # in graph 2, the graphs are likely not comparable.
        summary5 = summary.model_copy(deep=True)
        summary5.datasets["add_dataset3"] = summary.datasets["add_dataset3"].model_copy(
            deep=True,
            update={
                "producer": "task0",
                "n_visible": 0,
                "n_shadowed": 0,
                "n_predicted_only": 0,
                "n_expected": 1,
                "cursed_datasets": [],
                "unsuccessful_datasets": [
                    {"instrument": "INSTR", "detector": 0},
                ],
                "n_cursed": 0,
                "n_unsuccessful": 1,
            },
        )
        with self.assertLogs("lsst.pipe.base", level=lsst.utils.logging.VERBOSE) as warning_logs:
            Summary.aggregate([summary, summary5])
            self.assertIn(
                "WARNING:lsst.pipe.base.quantum_provenance_graph:Producer for dataset type is not consistent"
                ": 'task2' != 'task0'.",
                warning_logs.output[0],
            )
            self.assertIn(
                "WARNING:lsst.pipe.base.quantum_provenance_graph:Ignoring 'task0'.", warning_logs.output[1]
            )

        # Next up, we're going to try to aggregate summary with a dictionary
        # and then with some garbage. Neither of these should work!
        with self.assertRaises(AttributeError):
            Summary.aggregate(
                [
                    summary,
                    {
                        "tasks": {
                            "task0": {
                                "n_successful": 0,
                                "n_blocked": 0,
                                "n_unknown": 1,
                                "n_expected": 1,
                                "failed_quanta": [],
                                "recovered_quanta": [],
                                "wonky_quanta": [],
                                "n_wonky": 0,
                                "n_failed": 0,
                            },
                            "datasets": {
                                "add_dataset1": {
                                    "producer": "task0",
                                    "n_visible": 0,
                                    "n_shadowed": 0,
                                    "n_predicted_only": 0,
                                    "n_expected": 1,
                                    "cursed_datasets": [],
                                    "unsuccessful_datasets": [{"instrument": "INSTR", "detector": 0}],
                                    "n_cursed": 0,
                                    "n_unsuccessful": 1,
                                },
                                "add2_dataset1": {
                                    "producer": "task0",
                                    "n_visible": 0,
                                    "n_shadowed": 0,
                                    "n_predicted_only": 0,
                                    "n_expected": 1,
                                    "cursed_datasets": [],
                                    "unsuccessful_datasets": [{"instrument": "INSTR", "detector": 0}],
                                    "n_cursed": 0,
                                    "n_unsuccessful": 1,
                                },
                                "task0_metadata": {
                                    "producer": "task0",
                                    "n_visible": 0,
                                    "n_shadowed": 0,
                                    "n_predicted_only": 0,
                                    "n_expected": 1,
                                    "cursed_datasets": [],
                                    "unsuccessful_datasets": [{"instrument": "INSTR", "detector": 0}],
                                    "n_cursed": 0,
                                    "n_unsuccessful": 1,
                                },
                                "task0_log": {
                                    "producer": "task0",
                                    "n_visible": 0,
                                    "n_shadowed": 0,
                                    "n_predicted_only": 0,
                                    "n_expected": 1,
                                    "cursed_datasets": [],
                                    "unsuccessful_datasets": [{"instrument": "INSTR", "detector": 0}],
                                    "n_cursed": 0,
                                    "n_unsuccessful": 1,
                                },
                            },
                        }
                    },
                ]
            )
            Summary.aggregate([summary, []])
            Summary.aggregate([summary, "some_garbage"])



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = (
    "SerializedDatasetTypeNode",
    "SerializedEdge",
    "SerializedPipelineGraph",
    "SerializedTaskInitNode",
    "SerializedTaskNode",
    "SerializedTaskSubset",
    "expect_not_none",
)

from collections.abc import Mapping
from typing import Any, TypeVar

import networkx
import pydantic

from lsst.daf.butler import DatasetType, DimensionConfig, DimensionGroup, DimensionUniverse

from .. import automatic_connection_constants as acc
from ._dataset_types import DatasetTypeNode
from ._edges import Edge, ReadEdge, WriteEdge
from ._exceptions import PipelineGraphReadError
from ._nodes import NodeKey, NodeType
from ._pipeline_graph import PipelineGraph
from ._task_subsets import StepDefinitions, TaskSubset
from ._tasks import TaskImportMode, TaskInitNode, TaskNode

_U = TypeVar("_U")

_IO_VERSION_INFO = (0, 0, 1)
"""Version tuple embedded in saved PipelineGraphs.
"""


def expect_not_none(value: _U | None, msg: str) -> _U:
    """Check that a value is not `None` and return it.

    Parameters
    ----------
    value : `~typing.Any`
        Value to check.
    msg : `str`
        Error message for the case where ``value is None``.

    Returns
    -------
    value : `typing.Any`
        Value, guaranteed not to be `None`.

    Raises
    ------
    PipelineGraphReadError
        Raised with ``msg`` if ``value is None``.
    """
    if value is None:
        raise PipelineGraphReadError(msg)
    return value


class SerializedEdge(pydantic.BaseModel):
    """Struct used to represent a serialized `Edge` in a `PipelineGraph`.

    All `ReadEdge` and `WriteEdge` state not included here is instead
    effectively serialized by the context in which a `SerializedEdge` appears
    (e.g. the keys of the nested dictionaries in which it serves as the value
    type).
    """

    dataset_type_name: str
    """Full dataset type name (including component)."""

    storage_class: str
    """Name of the storage class."""

    raw_dimensions: list[str]
    """Raw dimensions of the dataset type from the task connections."""

    is_calibration: bool = False
    """Whether this dataset type can be included in
    `~lsst.daf.butler.CollectionType.CALIBRATION` collections."""

    defer_query_constraint: bool = False
    """If `True`, by default do not include this dataset type's existence as a
    constraint on the initial data ID query in QuantumGraph generation."""

    @classmethod
    def serialize(cls, target: Edge) -> SerializedEdge:
        """Transform an `Edge` to a `SerializedEdge`.

        Parameters
        ----------
        target : `Edge`
            The object to serialize.

        Returns
        -------
        `SerializedEdge`
            Model transformed into something that can be serialized.
        """
        return SerializedEdge.model_construct(
            storage_class=target.storage_class_name,
            dataset_type_name=target.dataset_type_name,
            raw_dimensions=sorted(target.raw_dimensions),
            is_calibration=target.is_calibration,
            defer_query_constraint=getattr(target, "defer_query_constraint", False),
        )

    def deserialize_read_edge(
        self,
        task_key: NodeKey,
        connection_name: str,
        dataset_type_keys: Mapping[str, NodeKey],
        is_prerequisite: bool = False,
    ) -> ReadEdge:
        """Transform a `SerializedEdge` to a `ReadEdge`.

        Parameters
        ----------
        task_key : `NodeKey`
            Key for the task node this edge is connected to.
        connection_name : `str`
            Internal name for the connection as seen by the task.
        dataset_type_keys : `~collections.abc.Mapping` [`str`, `NodeKey`]
            Mapping of dataset type name to node key.
        is_prerequisite : `bool`, optional
            Whether this dataset must be present in the data repository prior
            to `QuantumGraph` generation.

        Returns
        -------
        `ReadEdge`
            Deserialized object.
        """
        parent_dataset_type_name, component = DatasetType.splitDatasetTypeName(self.dataset_type_name)
        return ReadEdge(
            dataset_type_key=dataset_type_keys[parent_dataset_type_name],
            task_key=task_key,
            storage_class_name=self.storage_class,
            is_prerequisite=is_prerequisite,
            component=component,
            connection_name=connection_name,
            is_calibration=self.is_calibration,
            defer_query_constraint=self.defer_query_constraint,
            raw_dimensions=frozenset(self.raw_dimensions),
        )

    def deserialize_write_edge(
        self,
        task_key: NodeKey,
        connection_name: str,
        dataset_type_keys: Mapping[str, NodeKey],
    ) -> WriteEdge:
        """Transform a `SerializedEdge` to a `WriteEdge`.

        Parameters
        ----------
        task_key : `NodeKey`
            Key for the task node this edge is connected to.
        connection_name : `str`
            Internal name for the connection as seen by the task.
        dataset_type_keys : `~collections.abc.Mapping` [`str`, `NodeKey`]
            Mapping of dataset type name to node key.

        Returns
        -------
        `WriteEdge`
            Deserialized object.
        """
        return WriteEdge(
            task_key=task_key,
            dataset_type_key=dataset_type_keys[self.dataset_type_name],
            storage_class_name=self.storage_class,
            connection_name=connection_name,
            is_calibration=self.is_calibration,
            raw_dimensions=frozenset(self.raw_dimensions),
        )


class SerializedTaskInitNode(pydantic.BaseModel):
    """Struct used to represent a serialized `TaskInitNode` in a
    `PipelineGraph`.

    The task label is serialized by the context in which a
    `SerializedTaskInitNode` appears (e.g. the keys of the nested dictionary
    in which it serves as the value type), and the task class name and config
    string are save with the corresponding `SerializedTaskNode`.
    """

    inputs: dict[str, SerializedEdge]
    """Mapping of serialized init-input edges, keyed by connection name."""

    outputs: dict[str, SerializedEdge]
    """Mapping of serialized init-output edges, keyed by connection name."""

    config_output: SerializedEdge
    """The serialized config init-output edge."""

    index: int | None = None
    """The index of this node in the sorted sequence of `PipelineGraph`.

    This is `None` if the `PipelineGraph` was not sorted when it was
    serialized.
    """

    @classmethod
    def serialize(cls, target: TaskInitNode) -> SerializedTaskInitNode:
        """Transform a `TaskInitNode` to a `SerializedTaskInitNode`.

        Parameters
        ----------
        target : `TaskInitNode`
            Object to be serialized.

        Returns
        -------
        `SerializedTaskInitNode`
            Model that can be serialized.
        """
        return cls.model_construct(
            inputs={
                connection_name: SerializedEdge.serialize(edge)
                for connection_name, edge in sorted(target.inputs.items())
            },
            outputs={
                connection_name: SerializedEdge.serialize(edge)
                for connection_name, edge in sorted(target.outputs.items())
            },
            config_output=SerializedEdge.serialize(target.config_output),
        )

    def deserialize(
        self,
        key: NodeKey,
        task_class_name: str,
        config_str: str,
        dataset_type_keys: Mapping[str, NodeKey],
    ) -> TaskInitNode:
        """Transform a `SerializedTaskInitNode` to a `TaskInitNode`.

        Parameters
        ----------
        key : `NodeKey`
            Key that identifies this node in internal and exported networkx
            graphs.
        task_class_name : `str`, optional
            Fully-qualified name of the task class.  Must be provided if
            ``imported_data`` is not.
        config_str : `str`, optional
            Configuration for the task as a string of override statements.
        dataset_type_keys : `~collections.abc.Mapping` [`str`, `NodeKey`]
            Mapping of dataset type name to node key.

        Returns
        -------
        `TaskInitNode`
            Deserialized object.
        """
        return TaskInitNode(
            key,
            inputs={
                connection_name: serialized_edge.deserialize_read_edge(
                    key, connection_name, dataset_type_keys
                )
                for connection_name, serialized_edge in self.inputs.items()
            },
            outputs={
                connection_name: serialized_edge.deserialize_write_edge(
                    key, connection_name, dataset_type_keys
                )
                for connection_name, serialized_edge in self.outputs.items()
            },
            config_output=self.config_output.deserialize_write_edge(
                key, acc.CONFIG_INIT_OUTPUT_CONNECTION_NAME, dataset_type_keys
            ),
            task_class_name=task_class_name,
            config_str=config_str,
        )


class SerializedTaskNode(pydantic.BaseModel):
    """Struct used to represent a serialized `TaskNode` in a `PipelineGraph`.

    The task label is serialized by the context in which a
    `SerializedTaskNode` appears (e.g. the keys of the nested dictionary in
    which it serves as the value type).
    """

    task_class: str
    """Fully-qualified name of the task class."""

    init: SerializedTaskInitNode
    """Serialized task initialization node."""

    config_str: str
    """Configuration for the task as a string of override statements."""

    prerequisite_inputs: dict[str, SerializedEdge]
    """Mapping of serialized prerequisiste input edges, keyed by connection
    name.
    """

    inputs: dict[str, SerializedEdge]
    """Mapping of serialized input edges, keyed by connection name."""

    outputs: dict[str, SerializedEdge]
    """Mapping of serialized output edges, keyed by connection name."""

    metadata_output: SerializedEdge
    """The serialized metadata output edge."""

    dimensions: list[str]
    """The task's dimensions, if they were resolved."""

    log_output: SerializedEdge | None = None
    """The serialized log output edge."""

    index: int | None = None
    """The index of this node in the sorted sequence of `PipelineGraph`.

    This is `None` if the `PipelineGraph` was not sorted when it was
    serialized.
    """

    @classmethod
    def serialize(cls, target: TaskNode) -> SerializedTaskNode:
        """Transform a `TaskNode` to a `SerializedTaskNode`.

        Parameters
        ----------
        target : `TaskNode`
            Object to be serialized.

        Returns
        -------
        `SerializedTaskNode`
            Object that can be serialized.
        """
        dimensions = list(target.raw_dimensions)
        dimensions.sort()
        return cls.model_construct(
            task_class=target.task_class_name,
            init=SerializedTaskInitNode.serialize(target.init),
            config_str=target.get_config_str(),
            dimensions=dimensions,
            prerequisite_inputs={
                connection_name: SerializedEdge.serialize(edge)
                for connection_name, edge in sorted(target.prerequisite_inputs.items())
            },
            inputs={
                connection_name: SerializedEdge.serialize(edge)
                for connection_name, edge in sorted(target.inputs.items())
            },
            outputs={
                connection_name: SerializedEdge.serialize(edge)
                for connection_name, edge in sorted(target.outputs.items())
            },
            metadata_output=SerializedEdge.serialize(target.metadata_output),
            log_output=(
                SerializedEdge.serialize(target.log_output) if target.log_output is not None else None
            ),
        )

    def deserialize(
        self,
        key: NodeKey,
        init_key: NodeKey,
        dataset_type_keys: Mapping[str, NodeKey],
        universe: DimensionUniverse | None,
    ) -> TaskNode:
        """Transform a `SerializedTaskNode` to a `TaskNode`.

        Parameters
        ----------
        key : `NodeKey`
            Key that identifies this node in internal and exported networkx
            graphs.
        init_key : `NodeKey`
            Key that identifies the init node in internal and exported networkx
            graphs.
        dataset_type_keys : `~collections.abc.Mapping` [`str`, `NodeKey`]
            Mapping of dataset type name to node key.
        universe : `~lsst.daf.butler.DimensionUniverse` or `None`
            The dimension universe.

        Returns
        -------
        `TaskNode`
            Deserialized object.
        """
        init = self.init.deserialize(
            init_key,
            task_class_name=self.task_class,
            config_str=expect_not_none(
                self.config_str, f"No serialized config file for task with label {key.name!r}."
            ),
            dataset_type_keys=dataset_type_keys,
        )
        inputs = {
            connection_name: serialized_edge.deserialize_read_edge(key, connection_name, dataset_type_keys)
            for connection_name, serialized_edge in self.inputs.items()
        }
        prerequisite_inputs = {
            connection_name: serialized_edge.deserialize_read_edge(
                key, connection_name, dataset_type_keys, is_prerequisite=True
            )
            for connection_name, serialized_edge in self.prerequisite_inputs.items()
        }
        outputs = {
            connection_name: serialized_edge.deserialize_write_edge(key, connection_name, dataset_type_keys)
            for connection_name, serialized_edge in self.outputs.items()
        }
        if (serialized_log_output := self.log_output) is not None:
            log_output = serialized_log_output.deserialize_write_edge(
                key, acc.LOG_OUTPUT_CONNECTION_NAME, dataset_type_keys
            )
        else:
            log_output = None
        metadata_output = self.metadata_output.deserialize_write_edge(
            key, acc.METADATA_OUTPUT_CONNECTION_NAME, dataset_type_keys
        )
        dimensions: frozenset[str] | DimensionGroup
        if universe is not None:
            dimensions = universe.conform(self.dimensions)
        else:
            dimensions = frozenset(self.dimensions)
        return TaskNode(
            key=key,
            init=init,
            inputs=inputs,
            prerequisite_inputs=prerequisite_inputs,
            outputs=outputs,
            log_output=log_output,
            metadata_output=metadata_output,
            dimensions=dimensions,
        )


class SerializedDatasetTypeNode(pydantic.BaseModel):
    """Struct used to represent a serialized `DatasetTypeNode` in a
    `PipelineGraph`.

    Unresolved dataset types are serialized as instances with at most the
    `index` attribute set, and are typically converted to JSON with pydantic's
    ``exclude_defaults=True`` option to keep this compact.

    The dataset typename is serialized by the context in which a
    `SerializedDatasetTypeNode` appears (e.g. the keys of the nested dictionary
    in which it serves as the value type).
    """

    dimensions: list[str] | None = None
    """Dimensions of the dataset type."""

    storage_class: str | None = None
    """Name of the storage class."""

    is_calibration: bool = False
    """Whether this dataset type is a calibration."""

    is_initial_query_constraint: bool = False
    """Whether this dataset type should be a query constraint during
    `QuantumGraph` generation."""

    is_prerequisite: bool = False
    """Whether datasets of this dataset type must exist in the input collection
    before `QuantumGraph` generation."""

    index: int | None = None
    """The index of this node in the sorted sequence of `PipelineGraph`.

    This is `None` if the `PipelineGraph` was not sorted when it was
    serialized.
    """

    @classmethod
    def serialize(cls, target: DatasetTypeNode | None) -> SerializedDatasetTypeNode:
        """Transform a `DatasetTypeNode` to a `SerializedDatasetTypeNode`.

        Parameters
        ----------
        target : `DatasetTypeNode` or `None`
            Object to serialize.

        Returns
        -------
        `SerializedDatasetTypeNode`
            Object in serializable form.
        """
        if target is None:
            return cls.model_construct()
        return cls.model_construct(
            dimensions=list(target.dataset_type.dimensions.names),
            storage_class=target.dataset_type.storageClass_name,
            is_calibration=target.dataset_type.isCalibration(),
            is_initial_query_constraint=target.is_initial_query_constraint,
            is_prerequisite=target.is_prerequisite,
        )

    def deserialize(
        self, key: NodeKey, xgraph: networkx.MultiDiGraph, universe: DimensionUniverse | None
    ) -> DatasetTypeNode | None:
        """Transform a `SerializedDatasetTypeNode` to a `DatasetTypeNode`.

        Parameters
        ----------
        key : `NodeKey`
            Key that identifies this node in internal and exported networkx
            graphs.
        xgraph : `networkx.MultiDiGraph`
            <unknown>.
        universe : `~lsst.daf.butler.DimensionUniverse` or `None`
            The dimension universe.

        Returns
        -------
        `DatasetTypeNode`
            Deserialized object.
        """
        if self.dimensions is not None:
            dataset_type = DatasetType(
                key.name,
                expect_not_none(
                    self.dimensions,
                    f"Serialized dataset type {key.name!r} has no dimensions.",
                ),
                storageClass=expect_not_none(
                    self.storage_class,
                    f"Serialized dataset type {key.name!r} has no storage class.",
                ),
                isCalibration=self.is_calibration,
                universe=expect_not_none(
                    universe,
                    f"Serialized dataset type {key.name!r} has dimensions, "
                    "but no dimension universe was stored.",
                ),
            )
            producer: str | None = None
            producing_edge: WriteEdge | None = None
            for _, _, producing_edge in xgraph.in_edges(key, data="instance"):
                assert producing_edge is not None, "Should only be None if we never loop."
                if producer is not None:
                    raise PipelineGraphReadError(
                        f"Serialized dataset type {key.name!r} is produced by both "
                        f"{producing_edge.task_label!r} and {producer!r} in resolved graph."
                    )
                producer = producing_edge.task_label
            consuming_edges = tuple(
                consuming_edge for _, _, consuming_edge in xgraph.in_edges(key, data="instance")
            )
            return DatasetTypeNode(
                dataset_type=dataset_type,
                is_prerequisite=self.is_prerequisite,
                is_initial_query_constraint=self.is_initial_query_constraint,
                producing_edge=producing_edge,
                consuming_edges=consuming_edges,
            )
        return None


class SerializedTaskSubset(pydantic.BaseModel):
    """Struct used to represent a serialized `TaskSubset` in a `PipelineGraph`.

    The subset label is serialized by the context in which a
    `SerializedDatasetTypeNode` appears (e.g. the keys of the nested dictionary
    in which it serves as the value type).
    """

    description: str
    """Description of the subset."""

    tasks: list[str]
    """Labels of tasks in the subset, sorted lexicographically for
    determinism.
    """

    dimensions: list[str]
    """Dimensions that can be used to divide this step's quanta into
    independent groups.
    """

    @classmethod
    def serialize(cls, target: TaskSubset) -> SerializedTaskSubset:
        """Transform a `TaskSubset` into a `SerializedTaskSubset`.

        Parameters
        ----------
        target : `TaskSubset`
            Object to serialize.

        Returns
        -------
        `SerializedTaskSubset`
            Object in serializable form.
        """
        dimensions = sorted(target._step_definitions._dimensions_by_label.get(target.label, ()))
        return cls.model_construct(
            description=target._description,
            tasks=sorted(target),
            dimensions=dimensions,
        )

    def deserialize_task_subset(
        self, label: str, xgraph: networkx.MultiDiGraph, steps: StepDefinitions
    ) -> TaskSubset:
        """Transform a `SerializedTaskSubset` into a `TaskSubset`.

        Parameters
        ----------
        label : `str`
            Subset label.
        xgraph : `networkx.MultiDiGraph`
            The under-construction networkx graph that backs the pipeline
            graph.
        steps : `StepDefinitions`
            Step definitions for the pipeline graph.  Modified in-place to
            set sharding dimension if this is a step.

        Returns
        -------
        `TaskSubset`
            Deserialized object.
        """
        members = set(self.tasks)
        if label in steps:
            steps.set_dimensions(label, self.dimensions)
        return TaskSubset(xgraph, label, members, self.description, steps)


class SerializedPipelineGraph(pydantic.BaseModel):
    """Struct used to represent a serialized `PipelineGraph`."""

    version: str = ".".join(str(v) for v in _IO_VERSION_INFO)
    """Serialization version."""

    description: str
    """Human-readable description of the pipeline."""

    tasks: dict[str, SerializedTaskNode] = pydantic.Field(default_factory=dict)
    """Mapping of serialized tasks, keyed by label."""

    dataset_types: dict[str, SerializedDatasetTypeNode] = pydantic.Field(default_factory=dict)
    """Mapping of serialized dataset types, keyed by parent dataset type name.
    """

    task_subsets: dict[str, SerializedTaskSubset] = pydantic.Field(default_factory=dict)
    """Mapping of task subsets, keyed by subset label."""

    dimensions: dict[str, Any] | None = None
    """Dimension universe configuration."""

    data_id: dict[str, Any] = pydantic.Field(default_factory=dict)
    """Data ID that constrains all quanta generated from this pipeline."""

    step_labels: list[str] = pydantic.Field(default_factory=list)
    """List of task subset labels that are steps."""

    steps_verified: bool
    """Whether the step definitions in this pipeline were checked for
    consistency with the task and subset definitions.
    """

    @classmethod
    def serialize(cls, target: PipelineGraph) -> SerializedPipelineGraph:
        """Transform a `PipelineGraph` into a `SerializedPipelineGraph`.

        Parameters
        ----------
        target : `PipelineGraph`
            Object to serialize.

        Returns
        -------
        `SerializedPipelineGraph`
            Object in serializable form.
        """
        result = SerializedPipelineGraph.model_construct(
            description=target.description,
            tasks={label: SerializedTaskNode.serialize(node) for label, node in target.tasks.items()},
            dataset_types={
                name: SerializedDatasetTypeNode().serialize(target.dataset_types.get_if_resolved(name))
                for name in target.dataset_types
            },
            task_subsets={
                label: SerializedTaskSubset.serialize(subset) for label, subset in target.task_subsets.items()
            },
            step_labels=list(target.steps),
            steps_verified=target.steps.verified,
            dimensions=target._universe.dimensionConfig.toDict() if target._universe is not None else None,
            data_id=target._raw_data_id,
        )
        if target._sorted_keys:
            for index, node_key in enumerate(target._sorted_keys):
                match node_key.node_type:
                    case NodeType.TASK:
                        result.tasks[node_key.name].index = index
                    case NodeType.DATASET_TYPE:
                        result.dataset_types[node_key.name].index = index
                    case NodeType.TASK_INIT:
                        result.tasks[node_key.name].init.index = index
        return result

    def deserialize(
        self,
        import_mode: TaskImportMode,
    ) -> PipelineGraph:
        """Transform a `SerializedPipelineGraph` into a `PipelineGraph`.

        Parameters
        ----------
        import_mode : `TaskImportMode`
            Import mode.

        Returns
        -------
        `PipelineGraph`
            Deserialized object.
        """
        universe: DimensionUniverse | None = None
        if self.dimensions is not None:
            universe = DimensionUniverse(
                config=DimensionConfig(
                    expect_not_none(
                        self.dimensions,
                        "Serialized pipeline graph has not been resolved; "
                        "load it is a MutablePipelineGraph instead.",
                    )
                )
            )
        xgraph = networkx.MultiDiGraph()
        sort_index_map: dict[int, NodeKey] = {}
        # Save the dataset type keys after the first time we make them - these
        # may be tiny objects, but it's still to have only one copy of each
        # value floating around the graph.
        dataset_type_keys: dict[str, NodeKey] = {}
        for dataset_type_name, serialized_dataset_type in self.dataset_types.items():
            dataset_type_key = NodeKey(NodeType.DATASET_TYPE, dataset_type_name)
            # We intentionally don't attach a DatasetTypeNode instance here
            # yet, since we need edges to do that and those are saved with
            # the tasks.
            xgraph.add_node(dataset_type_key, bipartite=NodeType.DATASET_TYPE.value)
            if serialized_dataset_type.index is not None:
                sort_index_map[serialized_dataset_type.index] = dataset_type_key
            dataset_type_keys[dataset_type_name] = dataset_type_key
        for task_label, serialized_task in self.tasks.items():
            task_key = NodeKey(NodeType.TASK, task_label)
            task_init_key = NodeKey(NodeType.TASK_INIT, task_label)
            task_node = serialized_task.deserialize(task_key, task_init_key, dataset_type_keys, universe)
            if serialized_task.index is not None:
                sort_index_map[serialized_task.index] = task_key
            if serialized_task.init.index is not None:
                sort_index_map[serialized_task.init.index] = task_init_key
            xgraph.add_node(task_key, instance=task_node, bipartite=NodeType.TASK.bipartite)
            xgraph.add_node(task_init_key, instance=task_node.init, bipartite=NodeType.TASK_INIT.bipartite)
            xgraph.add_edge(task_init_key, task_key, Edge.INIT_TO_TASK_NAME, instance=None)
            for read_edge in task_node.init.iter_all_inputs():
                xgraph.add_edge(
                    read_edge.dataset_type_key,
                    read_edge.task_key,
                    read_edge.connection_name,
                    instance=read_edge,
                )
            for write_edge in task_node.init.iter_all_outputs():
                xgraph.add_edge(
                    write_edge.task_key,
                    write_edge.dataset_type_key,
                    write_edge.connection_name,
                    instance=write_edge,
                )
            for read_edge in task_node.iter_all_inputs():
                xgraph.add_edge(
                    read_edge.dataset_type_key,
                    read_edge.task_key,
                    read_edge.connection_name,
                    instance=read_edge,
                )
            for write_edge in task_node.iter_all_outputs():
                xgraph.add_edge(
                    write_edge.task_key,
                    write_edge.dataset_type_key,
                    write_edge.connection_name,
                    instance=write_edge,
                )
        # Iterate over dataset types again to add instances.
        for dataset_type_name, serialized_dataset_type in self.dataset_types.items():
            dataset_type_key = dataset_type_keys[dataset_type_name]
            xgraph.nodes[dataset_type_key]["instance"] = serialized_dataset_type.deserialize(
                dataset_type_key, xgraph, universe
            )
        steps = StepDefinitions(
            universe,
            dimensions_by_label=dict.fromkeys(self.step_labels, frozenset()),
            verified=self.steps_verified,
        )
        result = PipelineGraph.__new__(PipelineGraph)
        result._init_from_args(
            xgraph,
            sorted_keys=[sort_index_map[i] for i in range(len(xgraph))] if sort_index_map else None,
            task_subsets={
                subset_label: serialized_subset.deserialize_task_subset(subset_label, xgraph, steps)
                for subset_label, serialized_subset in self.task_subsets.items()
            },
            description=self.description,
            step_definitions=steps,
            universe=universe,
            data_id=self.data_id,
        )
        result._import_and_configure(import_mode)
        return result



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Expressions that resolve to subsets of pipelines.

See :ref:`pipeline-graph-subset-expressions`.
"""

from __future__ import annotations

__all__ = (
    "DirectionNode",
    "IdentifierNode",
    "IntersectionNode",
    "Node",
    "NotNode",
    "UnionNode",
    "parse",
)

import dataclasses
import functools
from typing import TYPE_CHECKING, Any, Literal, TypeAlias

from lsst.daf.butler.registry.queries.expressions.parser.ply import lex, yacc

from ._exceptions import InvalidExpressionError

if TYPE_CHECKING:
    from lsst.daf.butler.registry.queries.expressions.parser.parserLex import LexToken
    from lsst.daf.butler.registry.queries.expressions.parser.parserYacc import YaccProduction


class _ParserLex:
    @classmethod
    def make_lexer(cls) -> Any:  # unspecified PLY type.
        return lex.lex(object=cls())

    tokens = (
        "IDENTIFIER",
        "LPAREN",
        "RPAREN",
        "NOT",
        "UNION",
        "INTERSECTION",
        "LT",
        "LE",
        "GT",
        "GE",
    )

    t_LPAREN = r"\("
    t_RPAREN = r"\)"
    t_NOT = "~"
    t_UNION = r"\|"
    t_INTERSECTION = "&"
    t_LT = "<"
    t_LE = "<="
    t_GT = ">"
    t_GE = ">="

    # Identifiers are alphanumeric, and may have a T:, D:, or S: prefix.
    def t_IDENTIFIER(self, t: LexToken) -> LexToken:
        r"""([TDS]:)?\w+"""
        t.type = "IDENTIFIER"
        return t

    # Ignore spaces and tables.
    t_ignore = " \t"

    def t_error(self, t: LexToken) -> LexToken:
        raise InvalidExpressionError(
            f"invalid token in expression near character {t.lexer.lexpos}: {t.value[0]!r}"
        )


class _ParserYacc:
    def __init__(self) -> None:
        self.parser = self._parser_factory()

    @staticmethod
    @functools.cache
    def _parser_factory() -> Any:  # unspecified PLY type.
        return yacc.yacc(module=_ParserYacc, write_tables=False, debug=False)

    def parse(self, input: str) -> Node:
        """Parse input expression and return the parsed tree object.

        Parameters
        ----------
        input : `str`
            Expression to parse.

        Returns
        -------
        node : `Node`
            Root of the parsed expression tree.
        """
        lexer = _ParserLex.make_lexer()
        tree = self.parser.parse(input=input, lexer=lexer)
        return tree

    tokens = _ParserLex.tokens[:]

    start = "expr"

    precedence = (
        ("left", "UNION"),
        ("left", "INTERSECTION"),
        ("right", "NOT", "LT", "LE", "GT", "GE"),
    )

    # Ruff wants 'noqa' on the doc line, pydocstyle wants it on the function.

    @classmethod
    def p_expr_union(cls, p: YaccProduction) -> None:  # noqa: D403
        """expr : expr UNION expr"""  # noqa: D403
        p[0] = UnionNode(lhs=p[1], rhs=p[3])

    @classmethod
    def p_expr_intersection(cls, p: YaccProduction) -> None:  # noqa: D403
        """expr : expr INTERSECTION expr"""  # noqa: D403
        p[0] = IntersectionNode(lhs=p[1], rhs=p[3])

    @classmethod
    def p_expr_not(cls, p: YaccProduction) -> None:  # noqa: D403
        """expr : NOT expr"""  # noqa: D403
        p[0] = NotNode(operand=p[2])

    @classmethod
    def p_expr_parens(cls, p: YaccProduction) -> None:  # noqa: D403
        """expr : LPAREN expr RPAREN"""  # noqa: D403
        p[0] = p[2]

    @classmethod
    def p_expr_inequality(cls, p: YaccProduction) -> None:  # noqa: D403
        """expr : LT identifier
        | LE identifier
        | GT identifier
        | GE identifier
        """  # noqa: D403
        p[0] = DirectionNode(operator=p[1], start=p[2])

    @classmethod
    def p_expr_identifier(cls, p: YaccProduction) -> None:  # noqa: D403
        """expr : identifier"""  # noqa: D403
        p[0] = p[1]

    @classmethod
    def p_identifier_qualified(cls, p: YaccProduction) -> None:  # noqa: D403, D401
        """identifier : IDENTIFIER"""  # noqa: D403, D401
        match p[1].split(":"):
            case [qualifier, label]:
                p[0] = IdentifierNode(qualifier=qualifier, label=label)
            case [label]:
                p[0] = IdentifierNode(qualifier=None, label=label)
            case _:  # pragma: no cover
                raise AssertionError("Unexpected identifier form.")

    @classmethod
    def p_error(cls, p: YaccProduction | None) -> None:
        if p is None:
            raise InvalidExpressionError("Expression ended unexpectedly.")
        else:
            raise InvalidExpressionError(f"Syntax error near character {p.lexpos}: {p.value!r}")


@dataclasses.dataclass
class IdentifierNode:
    """A node that corresponds to a task label, dataset type name, or labeled
    subset.
    """

    qualifier: Literal["T", "D", "S"] | None
    """Qualiifier that indicates whether this is a task (T), dataset type (T),
    or labeled subset (S).

    Unqualified identifiers (`None`) must resolve unambiguously.
    """

    label: str
    """Task label, dataset type name, or subset label."""


@dataclasses.dataclass
class DirectionNode:
    """A node that represents the ancestors or descendents of a task label or
    dataset type.
    """

    operator: Literal["<", ">", "<=", ">="]
    """Which direction to traverse the graph ('>' for descendents, '<' for
    ancestors), and whether to include the operand ('=') or not.
    """

    start: IdentifierNode
    """Node at which to start the DAG traversal."""


@dataclasses.dataclass
class NotNode:
    """A node that represents set inversion (including all elements not in the
    operand).
    """

    operand: Node
    """Node representing the set to invert."""


@dataclasses.dataclass
class UnionNode:
    """Node representing a set union."""

    lhs: Node
    rhs: Node


@dataclasses.dataclass
class IntersectionNode:
    """Node representing a set intersection."""

    lhs: Node
    rhs: Node


def parse(expression: str) -> Node:
    """Parse an expression into a `Node` tree.

    Parameters
    ----------
    expression : `str`
        String expression to parse.  See
        :ref:`pipeline-graph-subset-expressions`.

    Returns
    -------
    node
        Root node of the parsed expression tree.

    Raises
    ------
    InvalidExpressionError
        Raised if the expression could not be parsed.
    """
    return _ParserYacc().parse(expression)


Node: TypeAlias = IdentifierNode | DirectionNode | NotNode | UnionNode | IntersectionNode



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Module defining Pipeline class and related methods."""

from __future__ import annotations

__all__ = ["LabelSpecifier", "Pipeline", "TaskDef"]

import copy
import logging
import re
import urllib.parse

# -------------------------------
#  Imports of standard modules --
# -------------------------------
from collections.abc import Callable, Set
from dataclasses import dataclass
from types import MappingProxyType
from typing import TYPE_CHECKING, cast

# -----------------------------
#  Imports for other modules --
# -----------------------------
from lsst.daf.butler import DataCoordinate, DimensionUniverse, Registry
from lsst.resources import ResourcePath, ResourcePathExpression
from lsst.utils import doImportType
from lsst.utils.introspection import get_full_type_name

from . import automatic_connection_constants as acc
from . import pipeline_graph, pipelineIR
from ._instrument import Instrument as PipeBaseInstrument
from .config import PipelineTaskConfig
from .connections import PipelineTaskConnections
from .pipelineTask import PipelineTask

if TYPE_CHECKING:  # Imports needed only for type annotations; may be circular.
    from lsst.obs.base import Instrument
    from lsst.pex.config import Config

# ----------------------------------
#  Local non-exported definitions --
# ----------------------------------

_LOG = logging.getLogger(__name__)

# ------------------------
#  Exported definitions --
# ------------------------


@dataclass
class LabelSpecifier:
    """A structure to specify a subset of labels to load.

    This structure may contain a set of labels to be used in subsetting a
    pipeline, or a beginning and end point. Beginning or end may be empty, in
    which case the range will be a half open interval. Unlike python iteration
    bounds, end bounds are *INCLUDED*.

    There are multiple potential definitions of range-based slicing for graphs
    that are not a simple linear sequence.  The definition used here is the
    intersection of the tasks downstream of ``begin`` and the tasks upstream of
    ``end``, i.e. tasks with no dependency relationship to a bounding task are
    not included.
    """

    labels: set[str] | None = None
    begin: str | None = None
    end: str | None = None

    def __post_init__(self) -> None:
        if self.labels is not None and (self.begin or self.end):
            raise ValueError(
                "This struct can only be initialized with a labels set or a begin (and/or) end specifier"
            )


class TaskDef:
    """TaskDef is a collection of information about task needed by Pipeline.

    The information includes task name, configuration object and optional
    task class. This class is just a collection of attributes and it exposes
    all of them so that attributes could potentially be modified in place
    (e.g. if configuration needs extra overrides).

    Parameters
    ----------
    taskName : `str`, optional
        The fully-qualified `PipelineTask` class name.  If not provided,
        ``taskClass`` must be.
    config : `lsst.pipe.base.config.PipelineTaskConfig`, optional
        Instance of the configuration class corresponding to this task class,
        usually with all overrides applied. This config will be frozen.  If
        not provided, ``taskClass`` must be provided and
        ``taskClass.ConfigClass()`` will be used.
    taskClass : `type`, optional
        `PipelineTask` class object; if provided and ``taskName`` is as well,
        the caller guarantees that they are consistent.  If not provided,
        ``taskName`` is used to import the type.
    label : `str`, optional
        Task label, usually a short string unique in a pipeline.  If not
        provided, ``taskClass`` must be, and ``taskClass._DefaultName`` will
        be used.
    connections : `PipelineTaskConnections`, optional
        Object that describes the dataset types used by the task.  If not
        provided, one will be constructed from the given configuration.  If
        provided, it is assumed that ``config`` has already been validated
        and frozen.
    """

    def __init__(
        self,
        taskName: str | None = None,
        config: PipelineTaskConfig | None = None,
        taskClass: type[PipelineTask] | None = None,
        label: str | None = None,
        connections: PipelineTaskConnections | None = None,
    ):
        if taskName is None:
            if taskClass is None:
                raise ValueError("At least one of `taskName` and `taskClass` must be provided.")
            taskName = get_full_type_name(taskClass)
        elif taskClass is None:
            taskClass = doImportType(taskName)
        if config is None:
            if taskClass is None:
                raise ValueError("`taskClass` must be provided if `config` is not.")
            config = taskClass.ConfigClass()
        if label is None:
            if taskClass is None:
                raise ValueError("`taskClass` must be provided if `label` is not.")
            label = taskClass._DefaultName
        self.taskName = taskName
        if connections is None:
            # If we don't have connections yet, assume the config hasn't been
            # validated yet.
            try:
                config.validate()
            except Exception:
                _LOG.error("Configuration validation failed for task %s (%s)", label, taskName)
                raise
            config.freeze()
            connections = config.connections.ConnectionsClass(config=config)
        self.config = config
        self.taskClass = taskClass
        self.label = label
        self.connections = connections

    @property
    def configDatasetName(self) -> str:
        """Name of a dataset type for configuration of this task (`str`)."""
        return acc.CONFIG_INIT_OUTPUT_TEMPLATE.format(label=self.label)

    @property
    def metadataDatasetName(self) -> str:
        """Name of a dataset type for metadata of this task (`str`)."""
        return self.makeMetadataDatasetName(self.label)

    @classmethod
    def makeMetadataDatasetName(cls, label: str) -> str:
        """Construct the name of the dataset type for metadata for a task.

        Parameters
        ----------
        label : `str`
            Label for the task within its pipeline.

        Returns
        -------
        name : `str`
            Name of the task's metadata dataset type.
        """
        return acc.METADATA_OUTPUT_TEMPLATE.format(label=label)

    @property
    def logOutputDatasetName(self) -> str | None:
        """Name of a dataset type for log output from this task, `None` if
        logs are not to be saved (`str`).
        """
        if self.config.saveLogOutput:
            return acc.LOG_OUTPUT_TEMPLATE.format(label=self.label)
        else:
            return None

    def __str__(self) -> str:
        rep = "TaskDef(" + self.taskName
        if self.label:
            rep += ", label=" + self.label
        rep += ")"
        return rep

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TaskDef):
            return False
        # This does not consider equality of configs when determining equality
        # as config equality is a difficult thing to define. Should be updated
        # after DM-27847
        return self.taskClass == other.taskClass and self.label == other.label

    def __hash__(self) -> int:
        return hash((self.taskClass, self.label))

    @classmethod
    def _unreduce(cls, taskName: str, config: PipelineTaskConfig, label: str) -> TaskDef:
        """Unpickle pickle. Custom callable for unpickling.

        All arguments are forwarded directly to the constructor; this
        trampoline is only needed because ``__reduce__`` callables can't be
        called with keyword arguments.
        """
        return cls(taskName=taskName, config=config, label=label)

    def __reduce__(self) -> tuple[Callable[[str, PipelineTaskConfig, str], TaskDef], tuple[str, Config, str]]:
        return (self._unreduce, (self.taskName, self.config, self.label))


class Pipeline:
    """A `Pipeline` is a representation of a series of tasks to run, and the
    configuration for those tasks.

    Parameters
    ----------
    description : `str`
        A description of that this pipeline does.
    """

    PipelineSubsetCtrl = pipelineIR.PipelineSubsetCtrl

    def __init__(self, description: str):
        pipeline_dict = {"description": description, "tasks": {}}
        self._pipelineIR = pipelineIR.PipelineIR(pipeline_dict)

    @classmethod
    def fromFile(cls, filename: str) -> Pipeline:
        """Load a pipeline defined in a pipeline yaml file.

        Parameters
        ----------
        filename : `str`
            A path that points to a pipeline defined in yaml format. This
            filename may also supply additional labels to be used in
            subsetting the loaded Pipeline. These labels are separated from
            the path by a ``#``, and may be specified as a comma separated
            list, or a range denoted as beginning..end. Beginning or end may
            be empty, in which case the range will be a half open interval.
            Unlike python iteration bounds, end bounds are *INCLUDED*. Note
            that range based selection is not well defined for pipelines that
            are not linear in nature, and correct behavior is not guaranteed,
            or may vary from run to run.

        Returns
        -------
        pipeline: `Pipeline`
            The pipeline loaded from specified location with appropriate (if
            any) subsetting.

        Notes
        -----
        This method attempts to prune any contracts that contain labels which
        are not in the declared subset of labels. This pruning is done using a
        string based matching due to the nature of contracts and may prune more
        than it should.
        """
        return cls.from_uri(filename)

    @classmethod
    def from_uri(cls, uri: ResourcePathExpression) -> Pipeline:
        """Load a pipeline defined in a pipeline yaml file at a location
        specified by a URI.

        Parameters
        ----------
        uri : convertible to `~lsst.resources.ResourcePath`
            If a string is supplied this should be a URI path that points to a
            pipeline defined in yaml format, either as a direct path to the
            yaml file, or as a directory containing a ``pipeline.yaml`` file
            the form used by `write_to_uri` with ``expand=True``). This uri may
            also supply additional labels to be used in subsetting the loaded
            `Pipeline`. These labels are separated from the path by a ``#``,
            and may be specified as a comma separated list, or a range denoted
            as beginning..end. Beginning or end may be empty, in which case the
            range will be a half open interval. Unlike python iteration bounds,
            end bounds are *INCLUDED*. Note that range based selection is not
            well defined for pipelines that are not linear in nature, and
            correct behavior is not guaranteed, or may vary from run to run.
            The same specifiers can be used with a
            `~lsst.resources.ResourcePath` object, by being the sole contents
            in the fragments attribute.

        Returns
        -------
        pipeline : `Pipeline`
            The pipeline loaded from specified location with appropriate (if
            any) subsetting.

        Notes
        -----
        This method attempts to prune any contracts that contain labels which
        are not in the declared subset of labels. This pruning is done using a
        string based matching due to the nature of contracts and may prune more
        than it should.
        """
        # Split up the uri and any labels that were supplied
        uri, label_specifier = cls._parse_file_specifier(uri)
        pipeline: Pipeline = cls.fromIR(pipelineIR.PipelineIR.from_uri(uri))

        # If there are labels supplied, only keep those
        if label_specifier is not None:
            pipeline = pipeline.subsetFromLabels(label_specifier)
        return pipeline

    def subsetFromLabels(
        self,
        labelSpecifier: LabelSpecifier,
        subsetCtrl: pipelineIR.PipelineSubsetCtrl = PipelineSubsetCtrl.DROP,
    ) -> Pipeline:
        """Subset a pipeline to contain only labels specified in
        ``labelSpecifier``.

        Parameters
        ----------
        labelSpecifier : `labelSpecifier`
            Object containing labels that describes how to subset a pipeline.
        subsetCtrl : `PipelineSubsetCtrl`
            Control object which decides how subsets with missing labels are
            handled. Setting to `PipelineSubsetCtrl.DROP` (the default) will
            cause any subsets that have labels which are not in the set of all
            task labels to be dropped. Setting to `PipelineSubsetCtrl.EDIT`
            will cause the subset to instead be edited to remove the
            nonexistent label.

        Returns
        -------
        pipeline : `Pipeline`
            A new pipeline object that is a subset of the old pipeline.

        Raises
        ------
        ValueError
            Raised if there is an issue with specified labels

        Notes
        -----
        This method attempts to prune any contracts that contain labels which
        are not in the declared subset of labels. This pruning is done using a
        string based matching due to the nature of contracts and may prune more
        than it should.
        """
        # Labels supplied as a set
        if labelSpecifier.labels:
            labelSet = labelSpecifier.labels
        # Labels supplied as a range, first create a list of all the labels
        # in the pipeline sorted according to task dependency. Then only
        # keep labels that lie between the supplied bounds
        else:
            # Create a copy of the pipeline to use when assessing the label
            # ordering. Use a dict for fast searching while preserving order.
            # Remove contracts so they do not fail in the expansion step. This
            # is needed because a user may only configure the tasks they intend
            # to run, which may cause some contracts to fail if they will later
            # be dropped
            pipeline = copy.deepcopy(self)
            pipeline._pipelineIR.contracts = []
            graph = pipeline.to_graph()

            # Verify the bounds are in the labels
            if labelSpecifier.begin is not None and labelSpecifier.begin not in graph.tasks:
                raise ValueError(
                    f"Beginning of range subset, {labelSpecifier.begin}, not found in pipeline definition"
                )
            if labelSpecifier.end is not None and labelSpecifier.end not in graph.tasks:
                raise ValueError(
                    f"End of range subset, {labelSpecifier.end}, not found in pipeline definition"
                )

            labelSet = set(graph.tasks.between(labelSpecifier.begin, labelSpecifier.end))
        return Pipeline.fromIR(self._pipelineIR.subset_from_labels(labelSet, subsetCtrl))

    @staticmethod
    def _parse_file_specifier(uri: ResourcePathExpression) -> tuple[ResourcePath, LabelSpecifier | None]:
        """Split appart a uri and any possible label subsets"""
        if isinstance(uri, str):
            # This is to support legacy pipelines during transition
            uri, num_replace = re.subn("[:](?!\\/\\/)", "#", uri)
            if num_replace:
                raise ValueError(
                    f"The pipeline file {uri} seems to use the legacy :"
                    " to separate labels, please use # instead."
                )
            if uri.count("#") > 1:
                raise ValueError("Only one set of labels is allowed when specifying a pipeline to load")
        # Everything else can be converted directly to ResourcePath.
        uri = ResourcePath(uri)
        label_subset = uri.fragment or None

        specifier: LabelSpecifier | None
        if label_subset is not None:
            label_subset = urllib.parse.unquote(label_subset)
            args: dict[str, set[str] | str | None]
            # labels supplied as a list
            if "," in label_subset:
                if ".." in label_subset:
                    raise ValueError(
                        "Can only specify a list of labels or a range when loading a Pipeline, not both."
                    )
                args = {"labels": set(label_subset.split(","))}
            # labels supplied as a range
            elif ".." in label_subset:
                # Try to de-structure the labelSubset, this will fail if more
                # than one range is specified
                begin, end, *rest = label_subset.split("..")
                if rest:
                    raise ValueError("Only one range can be specified when loading a pipeline")
                args = {"begin": begin if begin else None, "end": end if end else None}
            # Assume anything else is a single label
            else:
                args = {"labels": {label_subset}}

            # MyPy doesn't like how cavalier kwarg construction is with types.
            specifier = LabelSpecifier(**args)  # type: ignore
        else:
            specifier = None

        return uri, specifier

    @classmethod
    def fromString(cls, pipeline_string: str) -> Pipeline:
        """Create a pipeline from string formatted as a pipeline document.

        Parameters
        ----------
        pipeline_string : `str`
            A string that is formatted according like a pipeline document.

        Returns
        -------
        pipeline: `Pipeline`
            The new pipeline.
        """
        pipeline = cls.fromIR(pipelineIR.PipelineIR.from_string(pipeline_string))
        return pipeline

    @classmethod
    def fromIR(cls, deserialized_pipeline: pipelineIR.PipelineIR) -> Pipeline:
        """Create a pipeline from an already created `PipelineIR` object.

        Parameters
        ----------
        deserialized_pipeline : `PipelineIR`
            An already created pipeline intermediate representation object.

        Returns
        -------
        pipeline: `Pipeline`
            The new pipeline.
        """
        pipeline = cls.__new__(cls)
        pipeline._pipelineIR = deserialized_pipeline
        return pipeline

    @classmethod
    def fromPipeline(cls, pipeline: Pipeline) -> Pipeline:
        """Create a new pipeline by copying an already existing `Pipeline`.

        Parameters
        ----------
        pipeline : `Pipeline`
            An already created pipeline intermediate representation object.

        Returns
        -------
        pipeline: `Pipeline`
           The new pipeline.
        """
        return cls.fromIR(copy.deepcopy(pipeline._pipelineIR))

    def __str__(self) -> str:
        return str(self._pipelineIR)

    def mergePipeline(self, pipeline: Pipeline) -> None:
        """Merge another in-memory `Pipeline` object into this one.

        This merges another pipeline into this object, as if it were declared
        in the import block of the yaml definition of this pipeline. This
        modifies this pipeline in place.

        Parameters
        ----------
        pipeline : `Pipeline`
            The `Pipeline` object that is to be merged into this object.
        """
        self._pipelineIR.merge_pipelines((pipeline._pipelineIR,))

    def addLabelToSubset(self, subset: str, label: str) -> None:
        """Add a task label from the specified subset.

        Parameters
        ----------
        subset : `str`
            The labeled subset to modify.
        label : `str`
            The task label to add to the specified subset.

        Raises
        ------
        ValueError
            Raised if the specified subset does not exist within the pipeline.
            Raised if the specified label does not exist within the pipeline.
        """
        if label not in self._pipelineIR.tasks:
            raise ValueError(f"Label {label} does not appear within the pipeline")
        if subset not in self._pipelineIR.labeled_subsets:
            raise ValueError(f"Subset {subset} does not appear within the pipeline")
        self._pipelineIR.labeled_subsets[subset].subset.add(label)

    def removeLabelFromSubset(self, subset: str, label: str) -> None:
        """Remove a task label from the specified subset.

        Parameters
        ----------
        subset : `str`
            The labeled subset to modify.
        label : `str`
            The task label to remove from the specified subset.

        Raises
        ------
        ValueError
            Raised if the specified subset does not exist in the pipeline.
            Raised if the specified label does not exist within the specified
            subset.
        """
        if subset not in self._pipelineIR.labeled_subsets:
            raise ValueError(f"Subset {subset} does not appear within the pipeline")
        if label not in self._pipelineIR.labeled_subsets[subset].subset:
            raise ValueError(f"Label {label} does not appear within the pipeline")
        self._pipelineIR.labeled_subsets[subset].subset.remove(label)

    def findSubsetsWithLabel(self, label: str) -> set[str]:
        """Find any subsets which may contain the specified label.

        This function returns the name of subsets which return the specified
        label. May return an empty set if there are no subsets, or no subsets
        containing the specified label.

        Parameters
        ----------
        label : `str`
            The task label to use in membership check.

        Returns
        -------
        subsets : `set` of `str`
            Returns a set (possibly empty) of subsets names which contain the
            specified label.

        Raises
        ------
        ValueError
            Raised if the specified label does not exist within this pipeline.
        """
        results = set()
        if label not in self._pipelineIR.tasks:
            raise ValueError(f"Label {label} does not appear within the pipeline")
        for subset in self._pipelineIR.labeled_subsets.values():
            if label in subset.subset:
                results.add(subset.label)
        return results

    @property
    def task_labels(self) -> Set[str]:
        """Labels of all tasks in the pipelines.

        For simple pipelines with no imports, iteration over this set will
        match the order in which tasks are defined in the pipeline file.  In
        all other cases the order is unspecified but deterministic.  It is not
        dependency-ordered (use ``to_graph().tasks.keys()`` for that).
        """
        return self._pipelineIR.tasks.keys()

    @property
    def subsets(self) -> MappingProxyType[str, set]:
        """Returns a `MappingProxyType` where the keys are the labels of
        labeled subsets in the `Pipeline` and the values are the set of task
        labels contained within that subset.
        """
        return MappingProxyType(
            {label: subsetIr.subset for label, subsetIr in self._pipelineIR.labeled_subsets.items()}
        )

    def addLabeledSubset(self, label: str, description: str, taskLabels: set[str]) -> None:
        """Add a new labeled subset to the `Pipeline`.

        Parameters
        ----------
        label : `str`
            The label to assign to the subset.
        description : `str`
            A description of what the subset is for.
        taskLabels : `set` [`str`]
            The set of task labels to be associated with the labeled subset.

        Raises
        ------
        ValueError
            Raised if label already exists in the `Pipeline`.
            Raised if a task label is not found within the `Pipeline`.
        """
        if label in self._pipelineIR.labeled_subsets.keys():
            raise ValueError(f"Subset label {label} is already found within the Pipeline")
        if extra := (taskLabels - self._pipelineIR.tasks.keys()):
            raise ValueError(f"Task labels {extra} were not found within the Pipeline")
        self._pipelineIR.labeled_subsets[label] = pipelineIR.LabeledSubset(label, taskLabels, description)

    def removeLabeledSubset(self, label: str) -> None:
        """Remove a labeled subset from the `Pipeline`.

        Parameters
        ----------
        label : `str`
            The label of the subset to remove from the `Pipeline`.

        Raises
        ------
        ValueError
            Raised if the label is not found within the `Pipeline`.
        """
        if label not in self._pipelineIR.labeled_subsets.keys():
            raise ValueError(f"Subset label {label} was not found in the pipeline")
        self._pipelineIR.labeled_subsets.pop(label)

    def addInstrument(self, instrument: Instrument | str) -> None:
        """Add an instrument to the pipeline, or replace an instrument that is
        already defined.

        Parameters
        ----------
        instrument : `~lsst.daf.butler.instrument.Instrument` or `str`
            Either a derived class object of a `lsst.daf.butler.instrument` or
            a string corresponding to a fully qualified
            `lsst.daf.butler.instrument` name.
        """
        if isinstance(instrument, str):
            pass
        else:
            # TODO: assume that this is a subclass of Instrument, no type
            # checking
            instrument = get_full_type_name(instrument)
        self._pipelineIR.instrument = instrument

    def getInstrument(self) -> str | None:
        """Get the instrument from the pipeline.

        Returns
        -------
        instrument : `str`, or None
            The fully qualified name of a `lsst.obs.base.Instrument` subclass,
            name, or None if the pipeline does not have an instrument.
        """
        return self._pipelineIR.instrument

    def get_data_id(self, universe: DimensionUniverse) -> DataCoordinate:
        """Return a data ID with all dimension constraints embedded in the
        pipeline.

        Parameters
        ----------
        universe : `lsst.daf.butler.DimensionUniverse`
            Object that defines all dimensions.

        Returns
        -------
        data_id : `lsst.daf.butler.DataCoordinate`
            Data ID with all dimension constraints embedded in the
            pipeline.
        """
        instrument_class_name = self._pipelineIR.instrument
        if instrument_class_name is not None:
            instrument_class = cast(PipeBaseInstrument, doImportType(instrument_class_name))
            if instrument_class is not None:
                return DataCoordinate.standardize(instrument=instrument_class.getName(), universe=universe)
        return DataCoordinate.make_empty(universe)

    def addTask(self, task: type[PipelineTask] | str, label: str) -> None:
        """Add a new task to the pipeline, or replace a task that is already
        associated with the supplied label.

        Parameters
        ----------
        task : `PipelineTask` or `str`
            Either a derived class object of a `PipelineTask` or a string
            corresponding to a fully qualified `PipelineTask` name.
        label : `str`
            A label that is used to identify the `PipelineTask` being added.
        """
        if isinstance(task, str):
            taskName = task
        elif issubclass(task, PipelineTask):
            taskName = get_full_type_name(task)
        else:
            raise ValueError(
                "task must be either a child class of PipelineTask or a string containing"
                " a fully qualified name to one"
            )
        if not label:
            # in some cases (with command line-generated pipeline) tasks can
            # be defined without label which is not acceptable, use task
            # _DefaultName in that case
            if isinstance(task, str):
                task_class = cast(PipelineTask, doImportType(task))
                label = task_class._DefaultName
        self._pipelineIR.tasks[label] = pipelineIR.TaskIR(label, taskName)

    def removeTask(self, label: str) -> None:
        """Remove a task from the pipeline.

        Parameters
        ----------
        label : `str`
            The label used to identify the task that is to be removed.

        Raises
        ------
        KeyError
            If no task with that label exists in the pipeline.
        """
        self._pipelineIR.tasks.pop(label)

    def addConfigOverride(self, label: str, key: str, value: object) -> None:
        """Apply single config override.

        Parameters
        ----------
        label : `str`
            Label of the task.
        key : `str`
            Fully-qualified field name.
        value : object
            Value to be given to a field.
        """
        self._addConfigImpl(label, pipelineIR.ConfigIR(rest={key: value}))

    def addConfigFile(self, label: str, filename: str) -> None:
        """Add overrides from a specified file.

        Parameters
        ----------
        label : `str`
            The label used to identify the task associated with config to
            modify.
        filename : `str`
            Path to the override file.
        """
        self._addConfigImpl(label, pipelineIR.ConfigIR(file=[filename]))

    def addConfigPython(self, label: str, pythonString: str) -> None:
        """Add Overrides by running a snippet of python code against a config.

        Parameters
        ----------
        label : `str`
            The label used to identity the task associated with config to
            modify.
        pythonString : `str`
            A string which is valid python code to be executed. This is done
            with config as the only local accessible value.
        """
        self._addConfigImpl(label, pipelineIR.ConfigIR(python=pythonString))

    def _addConfigImpl(self, label: str, newConfig: pipelineIR.ConfigIR) -> None:
        if label == "parameters":
            self._pipelineIR.parameters.mapping.update(newConfig.rest)
            if newConfig.file:
                raise ValueError("Setting parameters section with config file is not supported")
            if newConfig.python:
                raise ValueError("Setting parameters section using python block in unsupported")
            return
        if label not in self._pipelineIR.tasks:
            raise LookupError(f"There are no tasks labeled '{label}' in the pipeline")
        self._pipelineIR.tasks[label].add_or_update_config(newConfig)

    def write_to_uri(self, uri: ResourcePathExpression) -> None:
        """Write the pipeline to a file or directory.

        Parameters
        ----------
        uri : convertible to `~lsst.resources.ResourcePath`
            URI to write to; may have any scheme with
            `~lsst.resources.ResourcePath` write support or no scheme for a
            local file/directory.  Should have a ``.yaml`` extension.
        """
        self._pipelineIR.write_to_uri(uri)

    def to_graph(
        self, registry: Registry | None = None, visualization_only: bool = False
    ) -> pipeline_graph.PipelineGraph:
        """Construct a pipeline graph from this pipeline.

        Constructing a graph applies all configuration overrides, freezes all
        configuration, checks all contracts, and checks for dataset type
        consistency between tasks (as much as possible without access to a data
        repository).  It cannot be reversed.

        Parameters
        ----------
        registry : `lsst.daf.butler.Registry`, optional
            Data repository client.  If provided, the graph's dataset types
            and dimensions will be resolved (see `PipelineGraph.resolve`).
        visualization_only : `bool`, optional
            Resolve the graph as well as possible even when dimensions and
            storage classes cannot really be determined.  This can include
            using the ``universe.commonSkyPix`` as the assumed dimensions of
            connections that use the "skypix" placeholder and using "<UNKNOWN>"
            as a storage class name (which will fail if the storage class
            itself is ever actually loaded).

        Returns
        -------
        graph : `pipeline_graph.PipelineGraph`
            Representation of the pipeline as a graph.
        """
        instrument_class_name = self._pipelineIR.instrument
        data_id = {}
        if instrument_class_name is not None:
            instrument_class: type[Instrument] = doImportType(instrument_class_name)
            if instrument_class is not None:
                data_id["instrument"] = instrument_class.getName()
        graph = pipeline_graph.PipelineGraph(data_id=data_id)
        graph.description = self._pipelineIR.description
        for label in self._pipelineIR.tasks:
            self._add_task_to_graph(label, graph)
        if self._pipelineIR.contracts is not None:
            label_to_config = {x.label: x.config for x in graph.tasks.values()}
            for contract in self._pipelineIR.contracts:
                # execute this in its own line so it can raise a good error
                # message if there was problems with the eval
                success = eval(contract.contract, None, label_to_config)
                if not success:
                    extra_info = f": {contract.msg}" if contract.msg is not None else ""
                    raise pipelineIR.ContractError(
                        f"Contract(s) '{contract.contract}' were not satisfied{extra_info}"
                    )
        for label, subset in self._pipelineIR.labeled_subsets.items():
            graph.add_task_subset(
                label, subset.subset, subset.description if subset.description is not None else ""
            )
        for step_ir in self._pipelineIR.steps:
            graph.steps.append(step_ir.label, step_ir.dimensions)
        if registry is not None or visualization_only:
            graph.resolve(registry=registry, visualization_only=visualization_only)
        else:
            graph.sort()
        return graph

    def _add_task_to_graph(self, label: str, graph: pipeline_graph.PipelineGraph) -> None:
        """Add a single task from this pipeline to a pipeline graph that is
        under construction.

        Parameters
        ----------
        label : `str`
            Label for the task to be added.
        graph : `pipeline_graph.PipelineGraph`
            Graph to add the task to.
        """
        if (taskIR := self._pipelineIR.tasks.get(label)) is None:
            raise NameError(f"Label {label} does not appear in this pipeline")
        taskClass: type[PipelineTask] = doImportType(taskIR.klass)
        config = taskClass.ConfigClass()
        instrument: PipeBaseInstrument | None = None
        if (instrumentName := self._pipelineIR.instrument) is not None:
            instrument_cls: type = doImportType(instrumentName)
            instrument = instrument_cls()
        config.applyConfigOverrides(
            instrument,
            getattr(taskClass, "_DefaultName", ""),
            taskIR.config,
            self._pipelineIR.parameters,
            label,
        )
        graph.add_task(label, taskClass, config)

    def __len__(self) -> int:
        return len(self._pipelineIR.tasks)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Pipeline):
            return False
        elif self._pipelineIR == other._pipelineIR:
            # Shortcut: if the IR is the same, the expanded pipeline must be
            # the same as well.  But the converse is not true.
            return True
        else:
            # Compare as much as we can (task classes and their edges).
            if self.to_graph().diff_tasks(other.to_graph()):
                return False
        # After DM-27847, we should compare configuration here.
        raise NotImplementedError(
            "Pipelines cannot be compared because config instances cannot be compared; see DM-27847."
        )



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__all__ = ["ExecutionGraphFixup"]

from abc import ABC, abstractmethod

from .graph import QuantumGraph


class ExecutionGraphFixup(ABC):
    """Interface for classes which update quantum graphs before execution.

    Primary goal of this class is to modify quanta dependencies which may not
    be possible to reflect in a quantum graph using standard tools. One known
    use case for that is to guarantee particular execution order of visits in
    CI jobs for cases when outcome depends on the processing order of visits
    (e.g. AP association pipeline).

    Instances of this class receive pre-ordered sequence of quanta
    (`.QuantumGraph` instances) and they are allowed to modify quanta data in
    place, for example update ``dependencies`` field to add additional
    dependencies. Returned list of quanta will be re-ordered once again by the
    graph executor to reflect new dependencies.
    """

    @abstractmethod
    def fixupQuanta(self, graph: QuantumGraph) -> QuantumGraph:
        """Update quanta in a graph.

        Potentially anything in the graph could be changed if it does not
        break executor assumptions. If modifications result in a dependency
        cycle the executor will raise an exception.

        Parameters
        ----------
        graph : `.QuantumGraph`
            Quantum Graph that will be executed by the executor.

        Returns
        -------
        graph : `.QuantumGraph`
            Modified graph.
        """
        raise NotImplementedError



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This software is dual licensed under the GNU General Public License and also
# under a 3-clause BSD license. Recipients may choose which of these licenses
# to use; please see the files gpl-3.0.txt and/or bsd_license.txt,
# respectively.  If you choose the GPL option then the following text applies
# (but note that there is still no warranty even if you opt for BSD instead):
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import unittest
from collections.abc import Callable
from types import MappingProxyType

from lsst.pipe.base import PipelineTaskConfig, PipelineTaskConnections
from lsst.pipe.base.connectionTypes import Input, Output, PrerequisiteInput

# Keyword arguments for defining our lone test connection in its class.
DEFAULT_CONNECTION_KWARGS = dict(
    doc="field docs",
    name="unconfigured",
    dimensions=(),
    storageClass="Dummy",
)

# Keyword arguments for making our lone test connection post-configuration.
# We always rename the dataset type via configuration to make sure those
# changes are never dropped.
RENAMED_CONNECTION_KWARGS = DEFAULT_CONNECTION_KWARGS.copy()
RENAMED_CONNECTION_KWARGS["name"] = "configured"


class TestDynamicConnectionsClass(unittest.TestCase):
    """Test modifying connections in derived __init__ implementations."""

    def build_dynamic_connections(
        self, init_callback: Callable[[PipelineTaskConnections], None] = None
    ) -> PipelineTaskConnections:
        """Define and construct a connections class instance with a callback
        run in ``__init__``.

        Parameters
        ----------
        init_callback
            Callback to invoke with the `PipelineTaskConnections` instance
            as its only argument.  Return value is ignored.

        Returns
        -------
        connections : `PipelineTaskConnections`
            Constructed connections instance.
        """

        class ExampleConnections(PipelineTaskConnections, dimensions=()):
            the_connection = Input(**DEFAULT_CONNECTION_KWARGS)

            def __init__(self, config: ExampleConfig):
                # Calling super() is harmless but now unnecessary, so don't do
                # it to make sure that works.  Old code calling it is fine.
                if init_callback is not None:
                    init_callback(self)

        class ExampleConfig(PipelineTaskConfig, pipelineConnections=ExampleConnections):
            pass

        config = ExampleConfig()
        config.connections.the_connection = RENAMED_CONNECTION_KWARGS["name"]
        return ExampleConnections(config=config)

    def test_freeze_after_construction(self):
        connections = self.build_dynamic_connections()
        self.assertIsInstance(connections.dimensions, frozenset)
        self.assertIsInstance(connections.inputs, frozenset)
        self.assertIsInstance(connections.prerequisiteInputs, frozenset)
        self.assertIsInstance(connections.outputs, frozenset)
        self.assertIsInstance(connections.initInputs, frozenset)
        self.assertIsInstance(connections.initOutputs, frozenset)
        self.assertIsInstance(connections.allConnections, MappingProxyType)

    def test_change_attr_after_construction(self):
        connections = self.build_dynamic_connections()
        with self.assertRaises(TypeError):
            connections.the_connection = PrerequisiteInput(**RENAMED_CONNECTION_KWARGS)

    def test_delete_attr_after_construction(self):
        connections = self.build_dynamic_connections()
        with self.assertRaises(TypeError):
            del connections.the_connection

    def test_change_dimensions(self):
        def callback(instance):
            instance.dimensions = {"new", "dimensions"}
            instance.the_connection = Input(
                doc=instance.the_connection.doc,
                name=instance.the_connection.name,
                dimensions={"new", "dimensions"},
                storageClass=instance.the_connection.storageClass,
            )

        connections = self.build_dynamic_connections(callback)
        self.assertEqual(connections.dimensions, {"new", "dimensions"})
        self.assertIsInstance(connections.dimensions, frozenset)
        self.assertEqual(connections.inputs, {"the_connection"})
        self.assertEqual(connections.allConnections.keys(), {"the_connection"})
        updated_connection_kwargs = RENAMED_CONNECTION_KWARGS.copy()
        updated_connection_kwargs["dimensions"] = {"new", "dimensions"}
        self.assertEqual(connections.allConnections["the_connection"], Input(**updated_connection_kwargs))
        self.assertEqual(connections.the_connection, Input(**updated_connection_kwargs))

    def test_change_connection_type(self):
        def callback(instance):
            instance.the_connection = PrerequisiteInput(
                doc=instance.the_connection.doc,
                name=instance.the_connection.name,
                dimensions=instance.the_connection.dimensions,
                storageClass=instance.the_connection.storageClass,
            )

        connections = self.build_dynamic_connections(callback)
        self.assertEqual(connections.inputs, set())
        self.assertEqual(connections.prerequisiteInputs, {"the_connection"})
        self.assertEqual(connections.allConnections.keys(), {"the_connection"})
        self.assertEqual(
            connections.allConnections["the_connection"], PrerequisiteInput(**RENAMED_CONNECTION_KWARGS)
        )
        self.assertEqual(connections.the_connection, PrerequisiteInput(**RENAMED_CONNECTION_KWARGS))

    def test_change_connection_type_twice(self):
        def callback(instance):
            instance.the_connection = PrerequisiteInput(
                doc=instance.the_connection.doc,
                name=instance.the_connection.name,
                dimensions=instance.the_connection.dimensions,
                storageClass=instance.the_connection.storageClass,
            )
            instance.the_connection = Output(
                doc=instance.the_connection.doc,
                name=instance.the_connection.name,
                dimensions=instance.the_connection.dimensions,
                storageClass=instance.the_connection.storageClass,
            )

        connections = self.build_dynamic_connections(callback)
        self.assertEqual(connections.inputs, set())
        self.assertEqual(connections.prerequisiteInputs, set())
        self.assertEqual(connections.outputs, {"the_connection"})
        self.assertEqual(connections.allConnections.keys(), {"the_connection"})
        self.assertEqual(connections.allConnections["the_connection"], Output(**RENAMED_CONNECTION_KWARGS))
        self.assertEqual(connections.the_connection, Output(**RENAMED_CONNECTION_KWARGS))

    def test_remove_from_set(self):
        def callback(instance):
            instance.inputs.remove("the_connection")
            # We can't make this remove corresponding attribute or the entry in
            # allConnections *immediately* without using a custom set class for
            # 'inputs' etc, which we haven't bothered to do, because even that
            # wouldn't be enough to have additions to those sets update the
            # attributes and allConnections.  Instead updates to those happen
            # after __init__.

        connections = self.build_dynamic_connections(callback)
        self.assertEqual(connections.inputs, set())
        self.assertEqual(connections.allConnections, {})
        with self.assertRaises(AttributeError):
            connections.the_connection

    def test_delete_attr(self):
        def callback(instance):
            del instance.the_connection
            # This updates the corresponding entry from the inputs set and
            # the allConnections dict.
            self.assertEqual(instance.inputs, set())
            self.assertEqual(instance.allConnections, {})

        connections = self.build_dynamic_connections(callback)
        self.assertEqual(connections.inputs, set())
        self.assertEqual(connections.allConnections, {})
        with self.assertRaises(AttributeError):
            connections.the_connection

    def test_delete_attr_twice(self):
        def callback(instance):
            del instance.the_connection
            with self.assertRaises(AttributeError):
                del instance.the_connection

        self.build_dynamic_connections(callback)

    def test_change_connection_type_then_remove_from_set(self):
        def callback(instance):
            instance.the_connection = PrerequisiteInput(
                doc=instance.the_connection.doc,
                name=instance.the_connection.name,
                dimensions=instance.the_connection.dimensions,
                storageClass=instance.the_connection.storageClass,
            )
            instance.prerequisiteInputs.remove("the_connection")

        connections = self.build_dynamic_connections(callback)
        self.assertEqual(connections.inputs, set())
        self.assertEqual(connections.prerequisiteInputs, set())
        self.assertEqual(connections.allConnections, {})
        with self.assertRaises(AttributeError):
            connections.the_connection
        with self.assertRaises(KeyError):
            connections.allConnections["the_connection"]

    def test_change_connection_type_then_delete_attr(self):
        def callback(instance):
            instance.the_connection = PrerequisiteInput(
                doc=instance.the_connection.doc,
                name=instance.the_connection.name,
                dimensions=instance.the_connection.dimensions,
                storageClass=instance.the_connection.storageClass,
            )
            del instance.the_connection
            # This updates the corresponding entry from the inputs set and
            # the allConnections dict.
            self.assertEqual(instance.inputs, set())
            self.assertEqual(instance.prerequisiteInputs, set())
            self.assertEqual(instance.allConnections, {})
            with self.assertRaises(AttributeError):
                instance.the_connection
            with self.assertRaises(KeyError):
                instance.allConnections["the_connection"]

        connections = self.build_dynamic_connections(callback)
        self.assertEqual(connections.inputs, set())
        self.assertEqual(connections.prerequisiteInputs, set())
        self.assertEqual(connections.allConnections, {})
        with self.assertRaises(AttributeError):
            connections.the_connection
        with self.assertRaises(KeyError):
            connections.allConnections["the_connection"]

    def test_add_new_connection(self):
        new_connection = Output(
            name="new_dataset_type",
            doc="new connection_docs",
            storageClass="Dummy",
            dimensions=(),
        )

        def callback(instance):
            instance.new_connection = new_connection
            self.assertEqual(instance.outputs, {"new_connection"})
            self.assertEqual(instance.allConnections.keys(), {"new_connection", "the_connection"})
            self.assertIs(instance.new_connection, new_connection)
            self.assertIs(instance.allConnections["new_connection"], new_connection)

        connections = self.build_dynamic_connections(callback)
        self.assertEqual(connections.outputs, {"new_connection"})
        self.assertEqual(connections.allConnections.keys(), {"new_connection", "the_connection"})
        self.assertIs(connections.new_connection, new_connection)
        self.assertIs(connections.allConnections["new_connection"], new_connection)

    def test_add_and_change_new_connection(self):
        new_connection = Output(
            name="new_dataset_type",
            doc="new connection_docs",
            storageClass="Dummy",
            dimensions=(),
        )
        changed_connection = PrerequisiteInput(
            name="new_dataset_type",
            doc="new connection_docs",
            storageClass="Dummy",
            dimensions=(),
        )

        def callback(instance):
            instance.new_connection = new_connection
            self.assertEqual(instance.outputs, {"new_connection"})
            self.assertEqual(instance.allConnections.keys(), {"new_connection", "the_connection"})
            self.assertIs(instance.new_connection, new_connection)
            self.assertIs(instance.allConnections["new_connection"], new_connection)
            instance.new_connection = changed_connection
            self.assertEqual(instance.outputs, set())
            self.assertEqual(instance.allConnections.keys(), {"new_connection", "the_connection"})
            self.assertIs(instance.new_connection, changed_connection)
            self.assertIs(instance.allConnections["new_connection"], changed_connection)

        connections = self.build_dynamic_connections(callback)
        self.assertEqual(connections.outputs, set())
        self.assertEqual(connections.allConnections.keys(), {"new_connection", "the_connection"})
        self.assertIs(connections.new_connection, changed_connection)
        self.assertIs(connections.allConnections["new_connection"], changed_connection)

    def test_add_and_remove_new_connection(self):
        new_connection = Output(
            name="new_dataset_type",
            doc="new connection_docs",
            storageClass="Dummy",
            dimensions=(),
        )

        def callback(instance):
            instance.new_connection = new_connection
            self.assertEqual(instance.outputs, {"new_connection"})
            self.assertEqual(instance.allConnections.keys(), {"new_connection", "the_connection"})
            self.assertIs(instance.new_connection, new_connection)
            self.assertIs(instance.allConnections["new_connection"], new_connection)
            del instance.new_connection
            self.assertEqual(instance.outputs, set())
            self.assertEqual(instance.allConnections.keys(), {"the_connection"})
            with self.assertRaises(AttributeError):
                instance.new_connection
            with self.assertRaises(KeyError):
                instance.allConnections["new_connection"]

        connections = self.build_dynamic_connections(callback)
        self.assertEqual(connections.outputs, set())
        self.assertEqual(connections.allConnections.keys(), {"the_connection"})
        with self.assertRaises(AttributeError):
            connections.new_connection
        with self.assertRaises(KeyError):
            connections.allConnections["new_connection"]


if __name__ == "__main__":
    unittest.main()
